#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import json
import urllib.parse
import websockets
import realtime
import logging

import octobot_commons.logging as commons_logging
import octobot.community.supabase_backend.supabase_realtime_channel as supabase_realtime_channel


# disable websockets logger
logging.getLogger("websockets.client").setLevel(logging.WARNING)


class AuthenticatedSupabaseRealtimeSocket(realtime.Socket):
    def __init__(self, url: str, auto_reconnect: bool = False, params: dict = {}, hb_interval: int = 5):
        """
        from realtime.Socket docs:
        `Socket` is the abstraction for an actual socket connection that receives and 'reroutes' `Message` according to its `topic` and `event`.
        Socket-Channel has a 1-many relationship.
        Socket-Topic has a 1-many relationship.
        :param url: Websocket URL of the Realtime server. starts with `ws://` or `wss://`
        :param params: Optional parameters for connection.
        :param hb_interval: WS connection is kept alive by sending a heartbeat message. Optional, defaults to 5.
        """
        # local override of super().__init__() to avoid python 3.8 incompatibility
        self.url = url
        self.channels = {}
        self.connected = False
        self.params = params
        self.hb_interval = hb_interval
        self.kept_alive = False
        self.auto_reconnect = auto_reconnect

        self.logger = commons_logging.get_logger(self.__class__.__name__)
        self.closed = True
        self.pending_subscribe_callbacks = {}
        self.system_callbacks = {}
        self.listen_task = None
        self.ws_connection = None

    def set_channel(self, schema: str, table_name: str) -> supabase_realtime_channel.AuthenticatedSupabaseRealtimeChannel:
        topic = (
            f"realtime:{schema}"
            if table_name == "*"
            else f"realtime:{schema}:{table_name}"
        )
        chan = supabase_realtime_channel.AuthenticatedSupabaseRealtimeChannel(
            self, topic, schema, table_name, self.params
        )
        try:
            self.channels[topic].append(chan)
        except KeyError:
            self.channels[topic] = [chan]
        return chan

    def register_subscribe_callback(self, topic, callback):
        if topic not in self.pending_subscribe_callbacks:
            self.pending_subscribe_callbacks[topic] = [callback]
        else:
            self.pending_subscribe_callbacks[topic].append(callback)

    def register_system_callback(self, topic, callback):
        if topic not in self.system_callbacks:
            self.system_callbacks[topic] = [callback]
        else:
            self.system_callbacks[topic].append(callback)

    async def _listen(self) -> None:
        """
        local override to handle errors and async callbacks and methods
        from realtime.Socket docs:
        An infinite loop that keeps listening.
        :return: None
        """
        try:
            reconnect_delay = 0
            while True:
                try:
                    msg = await self.ws_connection.recv()
                    # success as a message is received: reset reconnect delay
                    reconnect_delay = 0
                    await self._on_message(msg)
                except websockets.ConnectionClosed as err:  # pylint: disable=no-member
                    if self.closed or not self.auto_reconnect:
                        # should not try to reconnect, exit loop iteration
                        self.logger.exception(err, True, f"Realtime connection closed.")
                        break
                    # not closed and should reconnect: try to reconnect
                    self.logger.info("The realtime connection with server closed, trying to reconnect...")
                    if not await self.aconnect():
                        # reconnect failed, retry in next loop (ConnectionClosed will be raised again)
                        # exponential reconnect delay, caped at 60 seconds
                        reconnect_delay = min((reconnect_delay + 1) * 2, 60)
                        self.logger.info(f"Realtime reconnect failed. Next attempt in {reconnect_delay} seconds.")
                        await asyncio.sleep(reconnect_delay)
                        continue
                    # reconnect success: resubscribe and listen again in next loop iteration
                    await self.subscribe_channels()
        except Exception as err:
            self.logger.exception(err, True, f"Unexpected error when listening to  realtime message: {err}")
            raise

    async def subscribe_channels(self):
        for channels in self.channels.values():
            for channel in channels:
                if not (channel.is_joined() or channel.is_joining()):
                    await channel.subscribe()

    def _set_closed_channel_states(self):
        for channels in self.channels.values():
            for channel in channels:
                channel.state = supabase_realtime_channel.CHANNEL_STATES.CLOSED

    async def _on_message(self, text_msg):
        try:
            msg = realtime.Message(**json.loads(text_msg))

            if msg.event == realtime.ChannelEvents.reply:
                for subscribe_cb in self.pending_subscribe_callbacks.get(msg.topic, []):
                    await subscribe_cb(self.get_subscribe_result(msg), msg)
                # skip next callbacks
                return

            if msg.event == 'system':
                for system_cb in self.system_callbacks.get(msg.topic, []):
                    await system_cb(msg)
                # skip next callbacks
                return

            for channel in self.channels.get(msg.topic, []):
                for cl in channel.listeners:
                    if cl.event in ["*", msg.event]:
                        await cl.callback(msg.payload)
        except Exception as err:
            self.logger.exception(err, True, f"Error when processing realtime message: {err}")
            raise

    def get_subscribe_result(self, message: realtime.Message):
        # inspired from https://github.com/supabase/realtime-js/blob/master/src/RealtimeChannel.ts#L220
        if message.payload.get("status", None) == "ok":
            return "SUBSCRIBED"
        if message.payload.get("status", None) == "error":
            return "CHANNEL_ERROR"
        if message.payload.get("status", None) == "timeout":
            return "TIMED_OUT"

    async def alisten(self):
        try:
            await asyncio.gather(self._listen(), self._keep_alive())
        except Exception as err:
            self.logger.exception(err, True, f"Unexpected exception in listen loop: {err}")

    async def aconnect(self):
        try:
            await self._close_ws_connection_if_any()
            self.closed = False
            ws_connection = await websockets.connect(f"{self.url}?{urllib.parse.urlencode(self.params)}")  # pylint: disable=no-member
            if self.closed:
                # was closed while connecting, don't keep this connection
                await ws_connection.close()
                return
            if ws_connection.open:
                self.logger.info("Realtime socket connected")
                self.ws_connection = ws_connection
                self.connected = True
                if self.listen_task is None or self.listen_task.done():
                    self.listen_task = asyncio.create_task(self.alisten())
                return True
        except (OSError, OSError, asyncio.TimeoutError, websockets.exceptions.InvalidHandshake) as err:
            self.logger.exception(err, True, f"Error when connecting to realtime websocket: {err}")
        except Exception as err:
            self.logger.exception(err, True, f"Unexpected error when connecting to realtime websocket: {err}")
        return False

    async def close(self):
        if not self.closed:
            self.logger.debug("closing realtime connection")
        self.closed = True
        if self.listen_task and not self.listen_task.done():
            self.listen_task.cancel()
        self.connected = False
        await self._close_ws_connection_if_any()

    async def _close_ws_connection_if_any(self):
        if self.ws_connection is not None:
            self._set_closed_channel_states()
            await self.ws_connection.close()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.


class StartupInfo:
    FORCED_PROFILE_URL = "forced_profile_url"
    SUBSCRIBED_PRODUCTS_URLS = "subscribed_products_urls"

    def __init__(self, forced_profile_url, subscribed_products_urls):
        self.forced_profile_url = forced_profile_url
        self.subscribed_products_urls = subscribed_products_urls

    @staticmethod
    def from_dict(data):
        return StartupInfo(
            data.get(StartupInfo.FORCED_PROFILE_URL),
            [
                url
                for url in data.get(StartupInfo.SUBSCRIBED_PRODUCTS_URLS, []) or []
                if url  # skip unset urls
            ]
        )

    def __str__(self):
        return f"forced_profile_url: {self.forced_profile_url}, " \
               f"subscribed_products_urls: {self.subscribed_products_urls}"



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import pytest_asyncio

import octobot.strategy_optimizer as strategy_optimizer
import octobot_trading.api as trading_api
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants
import octobot_evaluators.constants as evaluators_constants

import tentacles.Evaluator.Strategies as Strategies
import tentacles.Evaluator.TA as Evaluator
import tentacles.Trading.Mode as Mode


@pytest_asyncio.fixture
async def optimizer_inputs():
    tentacles_setup_config = tentacles_manager_api.create_tentacles_setup_config_with_tentacles(
        Mode.DailyTradingMode,
        Strategies.SimpleStrategyEvaluator,
        Evaluator.RSIMomentumEvaluator,
        Evaluator.DoubleMovingAverageTrendEvaluator
    )
    trading_mode = trading_api.get_activated_trading_mode(tentacles_setup_config)
    return tentacles_setup_config, trading_mode


def _get_optimizer_user_input_config(tentacle, user_input_name, enabled, value):
    return {
        get_ui_identifier(tentacle, user_input_name): {
            "enabled": enabled,
            "tentacle": tentacle.get_name(),
            "user_input": user_input_name,
            "value": value
        }
    }


def get_ui_identifier(tentacle, user_input_name):
    return f"{tentacle.get_name()}-{user_input_name}"


def _get_mocked_ui_config():
    return {
        key: val
        for element in [
            _get_optimizer_user_input_config(
                Evaluator.RSIMomentumEvaluator, "period_length", True,
                {
                    strategy_optimizer.StrategyDesignOptimizer.CONFIG_MAX: 15,
                    strategy_optimizer.StrategyDesignOptimizer.CONFIG_MIN: 3,
                    strategy_optimizer.StrategyDesignOptimizer.CONFIG_STEP: 2
                }
            ),
            _get_optimizer_user_input_config(
                Evaluator.RSIMomentumEvaluator, "constrained_risk", True,
                {
                    strategy_optimizer.StrategyDesignOptimizer.CONFIG_MAX: 2,
                    strategy_optimizer.StrategyDesignOptimizer.CONFIG_MIN: 1,
                    strategy_optimizer.StrategyDesignOptimizer.CONFIG_STEP: 1
                }
            ),
            _get_optimizer_user_input_config(
                Strategies.SimpleStrategyEvaluator, commons_constants.CONFIG_TENTACLES_REQUIRED_CANDLES_COUNT, False,
                {
                    strategy_optimizer.StrategyDesignOptimizer.CONFIG_MAX: 0,
                    strategy_optimizer.StrategyDesignOptimizer.CONFIG_MIN: 0,
                    strategy_optimizer.StrategyDesignOptimizer.CONFIG_STEP: 1
                }
            ),
            _get_optimizer_user_input_config(
                Strategies.SimpleStrategyEvaluator, evaluators_constants.STRATEGIES_REQUIRED_TIME_FRAME, True,
                [
                    commons_enums.TimeFrames.FIVE_MINUTES.value,
                    commons_enums.TimeFrames.ONE_HOUR.value,
                    commons_enums.TimeFrames.ONE_DAY.value,
                ]
            ),
            _get_optimizer_user_input_config(
                Strategies.SimpleStrategyEvaluator, Strategies.SimpleStrategyEvaluator.RE_EVAL_TA_ON_RT_OR_SOCIAL, True,
                [
                    False
                ]
            ),
        ]
        for key, val in element.items()
    }


def _get_filter_setting_content(role, value):
    return {
        strategy_optimizer.StrategyDesignOptimizer.CONFIG_ROLE: role,
        strategy_optimizer.StrategyDesignOptimizer.CONFIG_VALUE: value,
    }


def _get_filter_setting(left_value, right_value, flat_value, operator):
    return {
        "user_input_left_operand": _get_filter_setting_content(
            "user_input_left_operand",
            left_value
        ),
        "user_input_right_operand": _get_filter_setting_content(
            "user_input_right_operand",
            right_value
        ),
        "text_right_operand": _get_filter_setting_content(
            "text_right_operand",
            flat_value

        ),
        "operator": _get_filter_setting_content(
            strategy_optimizer.OptimizerFilter.OPERATOR_KEY,
            operator.value
        ),
    }


def _get_mocked_filters_settings():
    return [
        _get_filter_setting(get_ui_identifier(Evaluator.RSIMomentumEvaluator, "period_length"), "null", 5,
                            commons_enums.LogicalOperators.LOWER_THAN),
        _get_filter_setting(get_ui_identifier(Evaluator.RSIMomentumEvaluator, "period_length"), "null", 15,
                            commons_enums.LogicalOperators.HIGHER_OR_EQUAL_TO),
    ]


MOCKED_OPTIMIZER_CONFIG = {
    strategy_optimizer.StrategyDesignOptimizer.CONFIG_FILTER_SETTINGS: _get_mocked_filters_settings(),
    strategy_optimizer.StrategyDesignOptimizer.CONFIG_USER_INPUTS: _get_mocked_ui_config(),
}


def _get_mocked_run_ui(ui_name, tentacle, value):
    return {
        strategy_optimizer.StrategyDesignOptimizer.CONFIG_USER_INPUT: ui_name,
        strategy_optimizer.StrategyDesignOptimizer.CONFIG_TENTACLE: [tentacle.get_name()],
        strategy_optimizer.StrategyDesignOptimizer.CONFIG_VALUE: value
    }


def _get_mocked_run(period_length, constrained_risk, time_frame):
    return (
        _get_mocked_run_ui("period_length", Evaluator.RSIMomentumEvaluator, period_length),
        _get_mocked_run_ui("constrained_risk", Evaluator.RSIMomentumEvaluator, constrained_risk),
        _get_mocked_run_ui(evaluators_constants.STRATEGIES_REQUIRED_TIME_FRAME,
                           Strategies.SimpleStrategyEvaluator,
                           [time_frame.value]),
        _get_mocked_run_ui(Strategies.SimpleStrategyEvaluator.RE_EVAL_TA_ON_RT_OR_SOCIAL,
                           Strategies.SimpleStrategyEvaluator, False),
    )


MOCKED_RUNS = [
    _get_mocked_run(period_length_val, constrained_risk_val, time_frame_val)
    for period_length_val in (5, 7, 9, 11, 13)
    for constrained_risk_val in (1, 2)
    for time_frame_val in (commons_enums.TimeFrames.FIVE_MINUTES,
                           commons_enums.TimeFrames.ONE_HOUR,
                           commons_enums.TimeFrames.ONE_DAY)
]


EXPECTED_RUNS_FROM_MOCK = {
    index + 6: val  # index + 6 as the first 5 runs have been filtered
    for index, val in enumerate(MOCKED_RUNS)
}



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import traceback


class Error:
    """
    Error is the error data wrapper use to upload errors to the error server
    """

    def __init__(self, error: Exception, title: str, timestamp: float, metrics_id: str):
        self.error: Exception = error
        self.title: str = str(title)
        self.first_timestamp: float = timestamp
        self.last_timestamp: float = timestamp
        self.count: int = 1
        self.metrics_id: str = metrics_id
        self.type: str = self.error.__class__.__name__ if self.error else ""
        self.stacktrace: list = traceback.format_exception(
            type(self.error), value=self.error, tb=self.error.__traceback__
        )[1:] if self.error and isinstance(self.error, Exception) else []

    def to_dict(self) -> dict:
        """
        Return the dict serialization of self
        """
        return {
            "title": self.title,
            "type": self.type,
            "stacktrace": self.stacktrace,
            "firsttimestamp": self.first_timestamp,
            "lasttimestamp": self.last_timestamp,
            "count": self.count,
            "metricsid": self.metrics_id,
        }

    def is_equivalent(self, other) -> bool:
        return (self.error is other.error or
                (type(self.error) is type(other.error)
                 and self.error.args == other.error.args)) and \
               self.title == other.title and \
               self.metrics_id == other.metrics_id and \
               self.type == other.type and \
               self.stacktrace == other.stacktrace

    def merge_equivalent(self, other):
        self.count += other.count
        if other.last_timestamp > self.last_timestamp:
            self.last_timestamp = other.last_timestamp



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import abc
import copy

import octobot.backtesting as octobot_backtesting
import octobot.api as octobot_api
import octobot_commons.constants as commons_constants
import octobot_commons.logging as bot_logging
import octobot_commons.tentacles_management as tentacles_management
import octobot_commons.tests.test_config as test_config
import octobot_evaluators.evaluators as evaluators
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_tentacles_manager.constants as tentacles_manager_constants
import octobot_trading.api as trading_api
import tests.test_utils.config as test_utils_config
import tentacles.Evaluator.Strategies as Strategies
import tentacles.Trading.Mode as Mode

DEFAULT_SYMBOL = "ICX/BTC"
DATA_FILE_PATH = "tests/static/"


class AbstractStrategyTest(octobot_backtesting.AbstractBacktestingTest, abc.ABC):
    __metaclass__ = abc.ABCMeta

    def initialize(self, strategy_evaluator_class, trading_mode_class=Mode.DailyTradingMode, config=None):
        self.logger = bot_logging.get_logger(self.__class__.__name__)
        self.config = test_config.load_test_config() if config is None else config
        # remove fees
        self.config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_SIMULATOR_FEES][
            commons_constants.CONFIG_SIMULATOR_FEES_TAKER] = 0
        self.config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_SIMULATOR_FEES][
            commons_constants.CONFIG_SIMULATOR_FEES_MAKER] = 0
        self.strategy_evaluator_class = strategy_evaluator_class
        self.trading_mode_class = trading_mode_class
        self.tentacles_setup_config = test_utils_config.load_test_tentacles_config()
        self._update_tentacles_config(strategy_evaluator_class, trading_mode_class)

    def _handle_results(self, independent_backtesting, profitability):
        exchange_manager_ids = octobot_api.get_independent_backtesting_exchange_manager_ids(independent_backtesting)
        for exchange_manager in trading_api.get_exchange_managers_from_exchange_ids(exchange_manager_ids):
            _, run_profitability, _, market_average_profitability, _ = trading_api.get_profitability_stats(
                exchange_manager)
            actual = round(run_profitability, 3)
            # uncomment this print for building tests
            # print(f"results: rounded run profitability {actual} market profitability: {market_average_profitability}"
            #       f" expected: {profitability} [result: {actual ==  profitability}]")
            assert actual == profitability

    async def _run_backtesting_with_current_config(self, data_file_to_use):
        config_to_use = copy.deepcopy(self.config)
        independent_backtesting = octobot_api.create_independent_backtesting(
            config_to_use,
            self.tentacles_setup_config,
            [data_file_to_use],
            "",
            enforce_total_databases_max_size_after_run=False,
            enable_storage=False,
        )
        await octobot_api.initialize_and_run_independent_backtesting(independent_backtesting, log_errors=False)
        await octobot_api.join_independent_backtesting(independent_backtesting)
        return independent_backtesting

    def _update_tentacles_config(self, strategy_evaluator_class, trading_mode_class):
        default_evaluators = strategy_evaluator_class.get_default_evaluators(self.tentacles_setup_config)
        to_update_config = {}
        tentacles_activation = tentacles_manager_api.get_tentacles_activation(self.tentacles_setup_config)
        for tentacle_class_name in tentacles_activation[tentacles_manager_constants.TENTACLES_EVALUATOR_PATH]:
            if commons_constants.CONFIG_WILDCARD not in default_evaluators and tentacle_class_name in default_evaluators:
                to_update_config[tentacle_class_name] = True
            elif tentacles_management.get_class_from_string(tentacle_class_name, evaluators.StrategyEvaluator,
                                                            Strategies,
                                                            tentacles_management.evaluator_parent_inspection) is not None:
                to_update_config[tentacle_class_name] = False
            elif commons_constants.CONFIG_WILDCARD not in default_evaluators:
                to_update_config[tentacle_class_name] = False
        for tentacle_class_name in tentacles_activation[tentacles_manager_constants.TENTACLES_TRADING_PATH]:
            to_update_config[tentacle_class_name] = False
        # Add required elements if missing
        to_update_config.update({evaluator: True for evaluator in default_evaluators})
        to_update_config[strategy_evaluator_class.get_name()] = True
        to_update_config[trading_mode_class.get_name()] = True
        tentacles_manager_api.update_activation_configuration(self.tentacles_setup_config, to_update_config, False,
                                                              add_missing_elements=True)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import dataclasses
import octobot.community.supabase_backend.enums as enums


class CommunityPublicData:
    def __init__(self):
        self.products = _DataElement({}, False)

    def set_products(self, products):
        self.products.value = {product[enums.ProductKeys.ID.value]: product for product in products}
        self.products.fetched = True

    def get_product_slug(self, product_id):
        return self.products.value[product_id][enums.ProductKeys.SLUG.value]


@dataclasses.dataclass
class _DataElement:
    value: any
    fetched: bool



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import contextlib
import json

import octobot.constants as constants
import octobot.community.errors as errors
import octobot.community.identifiers_provider as identifiers_provider
import octobot.community.models.community_supports as community_supports
import octobot.community.models.startup_info as startup_info
import octobot.community.models.community_user_account as community_user_account
import octobot.community.models.community_public_data as community_public_data
import octobot.community.models.formatters as formatters
import octobot.community.supabase_backend as supabase_backend
import octobot.community.supabase_backend.enums as backend_enums
import octobot.community.feeds as community_feeds
import octobot_commons.constants as commons_constants
import octobot_commons.enums as commons_enums
import octobot_commons.authentication as authentication
import octobot_commons.configuration as commons_configuration
import octobot_trading.enums as trading_enums


def _bot_data_update(func):
    async def wrapper(*args, **kwargs):
        self = args[0]
        if not self.is_logged_in_and_has_selected_bot():
            self.logger.debug(f"Skipping {func.__name__} update: no user selected bot.")
            return
        try:
            self.logger.debug(f"bot_data_update: {func.__name__} initiated.")
            return await func(*args, **kwargs)
        except Exception as err:
            self.logger.exception(err, True, f"Error when calling {func.__name__} {err}")
        finally:
            self.logger.debug(f"bot_data_update: {func.__name__} completed.")
    return wrapper


class CommunityAuthentication(authentication.Authenticator):
    """
    Authentication utility
    """
    ALLOWED_TIME_DELAY = 1 * commons_constants.MINUTE_TO_SECONDS
    NEW_ACCOUNT_INITIALIZE_TIMEOUT = 1 * commons_constants.MINUTE_TO_SECONDS
    LOGIN_TIMEOUT = 20
    MAX_UPLOADED_TRADES_COUNT = 100
    BOT_NOT_FOUND_RETRY_DELAY = 1
    AUTHORIZATION_HEADER = "authorization"
    SESSION_HEADER = "X-Session"
    GQL_AUTHORIZATION_HEADER = "Authorization"

    def __init__(self, feed_url, config=None, backend_url=None, backend_key=None):
        super().__init__()
        self.feed_url = feed_url
        self.backend_url = backend_url or identifiers_provider.IdentifiersProvider.BACKEND_URL
        self.backend_key = backend_key or identifiers_provider.IdentifiersProvider.BACKEND_KEY
        self.configuration_storage = supabase_backend.SyncConfigurationStorage(config)
        self.supabase_client = self._create_client()
        self.user_account = community_user_account.CommunityUserAccount()
        self.public_data = community_public_data.CommunityPublicData()
        self._community_feed = None

        self.initialized_event = None
        self._login_completed = None
        self._startup_info = None

        self._fetch_account_task = None

    @staticmethod
    def create(configuration: commons_configuration.Configuration, **kwargs):
        return CommunityAuthentication.instance(
            None,
            config=configuration,
            **kwargs,
        )

    def update(self, configuration: commons_configuration.Configuration):
        self.configuration_storage.configuration = configuration

    def get_logged_in_email(self):
        if self.user_account.has_user_data():
            return self.user_account.get_email()
        raise authentication.AuthenticationRequired()

    def get_packages(self):
        try:
            #TODO
            return []
        except json.JSONDecodeError:
            return []

    def is_feed_connected(self):
        return self._community_feed is not None and self._community_feed.is_connected_to_remote_feed()

    def get_feed_last_message_time(self):
        if self._community_feed is None:
            return None
        return self._community_feed.last_message_time

    def has_filled_form(self, form_id):
        if not self.user_account.has_user_data():
            raise authentication.AuthenticationRequired()
        return form_id in self.user_account.get_filled_forms_ids()

    async def register_filled_form(self, form_id):
        if self.has_filled_form(form_id):
            return
        updated_filled_forms = self.user_account.get_filled_forms_ids()
        updated_filled_forms.append(form_id)
        await self._update_account_metadata({
            self.user_account.FILLED_FORMS: updated_filled_forms
        })

    def get_user_id(self):
        if not self.user_account.has_user_data():
            raise authentication.AuthenticationRequired()
        return self.user_account.get_user_id()

    async def get_deployment_url(self):
        deployment_url_data = await self.supabase_client.fetch_deployment_url(
            self.user_account.get_selected_bot_deployment_id()
        )
        return self.user_account.get_bot_deployment_url(deployment_url_data)

    async def get_gpt_signal(
        self, exchange: str, symbol: str, time_frame: commons_enums.TimeFrames, candle_open_time: float, version: str
    ) -> str:
        return await self.supabase_client.fetch_gpt_signal(exchange, symbol, time_frame, candle_open_time, version)

    async def get_gpt_signals_history(
        self, exchange: str, symbol: str, time_frame: commons_enums.TimeFrames,
        first_open_time: float, last_open_time: float, version: str
    ) -> dict:
        return await self.supabase_client.fetch_gpt_signals_history(
            exchange, symbol, time_frame, first_open_time, last_open_time, version
        )

    def get_is_signal_receiver(self):
        if self._community_feed is None:
            return False
        return self._community_feed.is_signal_receiver

    def get_is_signal_emitter(self):
        if self._community_feed is None:
            return False
        return self._community_feed.is_signal_emitter

    def get_signal_community_url(self, signal_identifier):
        try:
            slug = self.public_data.get_product_slug(signal_identifier)
            return f"{identifiers_provider.IdentifiersProvider.COMMUNITY_URL}/strategies/{slug}"
        except KeyError:
            return identifiers_provider.IdentifiersProvider.COMMUNITY_URL

    async def update_supports(self):
        def _supports_mock():
            return {
                "data": {
                    "attributes": {
                        "support_role": self.user_account.get_support_role()
                    }
                }
            }
        self._update_supports(200, _supports_mock())
        # TODO use real support fetch when implemented

    def _create_client(self):
        return supabase_backend.CommunitySupabaseClient(
            self.backend_url,
            self.backend_key,
            self.configuration_storage
        )

    async def _ensure_async_loop(self):
        # elements should be bound to the current loop
        if not self.is_using_the_current_loop():
            if self._login_completed is not None:
                should_set = self._login_completed.is_set()
                self._login_completed = asyncio.Event()
                if should_set:
                    self._login_completed.set()
            # changed event loop: restart client
            await self.supabase_client.close()
            self.user_account.flush()
            self.supabase_client = self._create_client()

    def is_using_the_current_loop(self):
        return self.supabase_client.event_loop is None \
            or self.supabase_client.event_loop is asyncio.get_event_loop()

    def is_initialized(self):
        return self.initialized_event is not None and self.initialized_event.is_set()

    def init_account(self):
        self._fetch_account_task = asyncio.create_task(self._initialize_account())

    async def async_init_account(self):
        self.init_account()
        await self._fetch_account_task

    async def _create_community_feed_if_necessary(self) -> bool:
        if self._community_feed is None:
            self._community_feed = community_feeds.community_feed_factory(
                self.feed_url,
                self,
                constants.COMMUNITY_FEED_DEFAULT_TYPE
            )
            return True
        return False

    async def _ensure_init_community_feed(self):
        await self._create_community_feed_if_necessary()
        if not self._community_feed.is_connected() and self._community_feed.can_connect():
            if self.initialized_event is not None and not self.initialized_event.is_set():
                await asyncio.wait_for(self.initialized_event.wait(), self.LOGIN_TIMEOUT)
        await self._community_feed.start()

    async def register_feed_callback(self, channel_type: commons_enums.CommunityChannelTypes, callback, identifier=None):
        try:
            await self._ensure_init_community_feed()
            await self._community_feed.register_feed_callback(channel_type, callback, identifier=identifier)
        except errors.BotError as e:
            self.logger.error(f"Impossible to connect to community signals: {e}")

    async def send(self, message, channel_type, identifier=None):
        """
        Sends a message
        """
        await self._ensure_init_community_feed()
        await self._community_feed.send(message, channel_type, identifier)

    async def wait_for_login_if_processing(self):
        if self._login_completed is not None and not self._login_completed.is_set():
            # ensure login details have been fetched
            await asyncio.wait_for(self._login_completed.wait(), self.LOGIN_TIMEOUT)

    def can_authenticate(self):
        return bool(
            identifiers_provider.IdentifiersProvider.BACKEND_URL
            and identifiers_provider.IdentifiersProvider.BACKEND_KEY
        )

    def must_be_authenticated_through_authenticator(self):
        return constants.IS_CLOUD_ENV

    async def login(self, email, password, password_token=None, minimal=False):
        self._ensure_email(email)
        self._ensure_community_url()
        self._reset_tokens()
        with self._login_process():
            if password_token:
                await self.supabase_client.sign_in_with_otp_token(password_token)
            else:
                await self.supabase_client.sign_in(email, password)
            await self._on_account_updated()
        if self.is_logged_in():
            await self.on_signed_in(minimal=minimal)

    async def register(self, email, password):
        if self.must_be_authenticated_through_authenticator():
            raise authentication.AuthenticationError("Creating a new account is not authorized on this environment.")
        # always logout before creating a new account
        self.logout()
        self._ensure_community_url()
        with self._login_process():
            await self.supabase_client.sign_up(email, password)
            await self._on_account_updated()
        if self.is_logged_in():
            await self.on_signed_in()

    async def on_signed_in(self, minimal=False):
        self.logger.info(f"Signed in as {self.get_logged_in_email()}")
        await self._initialize_account(minimal=minimal)

    async def _update_account_metadata(self, metadata_update):
        await self.supabase_client.update_metadata(metadata_update)
        await self._on_account_updated()

    async def update_selected_bot(self):
        self.user_account.flush_bot_details()
        await self._load_bot_if_selected()
        if not self.user_account.has_selected_bot_data():
            self.logger.info(self.user_account.NO_SELECTED_BOT_DESC)

    async def _load_bot_if_selected(self):
        # 1. use user selected bot id if any
        if saved_bot_id := self._get_saved_bot_id():
            try:
                await self.select_bot(saved_bot_id)
                return
            except errors.BotNotFoundError as e:
                # proceed to 2.
                self.logger.warning(str(e))
        # 2. fetch all user bots and create one if none, otherwise ask use for which one to use
        await self.load_user_bots()
        if len(self.user_account.get_all_user_bots_raw_data()) == 0:
            await self.select_bot(
                self.user_account.get_bot_id(
                    await self.create_new_bot()
                )
            )
        # more than one possible bot, can't auto-select one

    async def create_new_bot(self):
        deployment_type = backend_enums.DeploymentTypes.CLOUD if constants.IS_CLOUD_ENV \
            else backend_enums.DeploymentTypes.SELF_HOSTED
        return await self.supabase_client.create_bot(deployment_type)

    async def select_bot(self, bot_id):
        fetched_bot = await self.supabase_client.fetch_bot(bot_id)
        self.user_account.set_selected_bot_raw_data(fetched_bot)
        bot_name = self.user_account.get_bot_name_or_id(self.user_account.get_selected_bot_raw_data())
        self.logger.debug(f"Selected bot '{bot_name}'")
        self.user_account.bot_id = bot_id
        self._save_bot_id(self.user_account.bot_id)
        await self.on_new_bot_select()

    async def load_user_bots(self):
        self.user_account.set_all_user_bots_raw_data(
            self._get_self_hosted_bots(
                await self.supabase_client.fetch_bots()
            )
        )

    async def get_startup_info(self):
        if self._startup_info is None:
            self.user_account.ensure_selected_bot_id()
            self._startup_info = startup_info.StartupInfo.from_dict(
                await self.supabase_client.fetch_startup_info(
                    self.user_account.bot_id
                )
            )
        return self._startup_info

    async def get_subscribed_profile_urls(self):
        return await self.supabase_client.fetch_subscribed_products_urls()

    async def get_current_bot_products_subscription(self) -> dict:
        self.user_account.ensure_selected_bot_id()
        return await self.supabase_client.fetch_bot_products_subscription(
            self.user_account.get_selected_bot_deployment_id()
        )

    def is_logged_in_and_has_selected_bot(self):
        return (self.supabase_client.is_admin or self.is_logged_in()) and self.user_account.bot_id is not None

    async def refresh_selected_bot(self):
        self.user_account.set_selected_bot_raw_data(
            await self.supabase_client.fetch_bot(self.user_account.bot_id)
        )

    def _get_self_hosted_bots(self, bots):
        return [
            bot
            for bot in bots
            if self.user_account.is_self_hosted(bot)
        ]

    async def on_new_bot_select(self):
        pass

    def logout(self):
        """
        logout and remove saved auth details
        Warning: also call stop_feeds if feeds have to be stopped (not done here to keep method sync)
        """
        self.supabase_client.sign_out()
        self._reset_tokens()
        self.remove_login_detail()
        if self._community_feed is not None:
            self._community_feed.remove_device_details()

    def is_logged_in(self):
        return bool(self.supabase_client.is_signed_in() and self.user_account.has_user_data())

    def has_login_info(self):
        return self.supabase_client.has_login_info()

    def remove_login_detail(self):
        self.user_account.flush()
        self._reset_login_token()
        self._save_bot_id("")
        self.logger.debug("Removed community login data")

    async def stop(self):
        self.logger.debug("Stopping ...")
        if self._fetch_account_task is not None and not self._fetch_account_task.done():
            self._fetch_account_task.cancel()
        await self.supabase_client.close()
        self.logger.debug("Stopped")

    def _update_supports(self, resp_status, json_data):
        if resp_status == 200:
            self.user_account.supports = community_supports.CommunitySupports.from_community_dict(json_data)
            self.logger.debug(f"Fetched supports data.")
        else:
            self.logger.error(f"Error when fetching community support, "
                              f"error code: {resp_status}")

    @contextlib.contextmanager
    def _login_process(self):
        try:
            if self._login_completed is None:
                self._login_completed = asyncio.Event()
            self._login_completed.clear()
            yield
        finally:
            if not self._login_completed.is_set():
                self._login_completed.set()

    async def _initialize_account(self, minimal=False):
        try:
            await self._ensure_async_loop()
            self.initialized_event = asyncio.Event()
            if not (self.is_logged_in() or await self._restore_previous_session()):
                return
            self._login_completed.set()
            if not minimal:
                await self._ensure_init_community_feed()
                await self.update_supports()
                await self.update_selected_bot()
                self.logger.debug(f"Fetched account data")
                await self.init_public_data()
        except authentication.UnavailableError as e:
            self.logger.exception(e, True, f"Error when fetching community supports, "
                                           f"please check your internet connection.")
        except Exception as e:
            self.logger.exception(e, True, f"Error when fetching community supports: {e}({e.__class__.__name__})")
        finally:
            self.initialized_event.set()

    async def init_public_data(self, reset=False):
        if reset or not self.public_data.products.fetched:
            self.public_data.set_products(await self.supabase_client.fetch_products())

    def _reset_login_token(self):
        if self.supabase_client is not None:
            self._save_value_in_config(self.supabase_client.auth._storage_key, "")

    def _save_bot_id(self, bot_id):
        self._save_value_in_config(constants.CONFIG_COMMUNITY_BOT_ID, bot_id)

    def _get_saved_bot_id(self):
        return constants.COMMUNITY_BOT_ID or self._get_value_in_config(constants.CONFIG_COMMUNITY_BOT_ID)

    def _save_value_in_config(self, key, value):
        self.configuration_storage.set_item(key, value)

    def _get_value_in_config(self, key):
        return self.configuration_storage.get_item(key)

    async def _restore_previous_session(self):
        with self._login_process():
            async with self._auth_handler():
                # will raise on failure
                self.supabase_client.restore_session()
                await self._on_account_updated()
                self.logger.info(f"Signed in as {self.get_logged_in_email()}")
        return self.is_logged_in()

    @contextlib.asynccontextmanager
    async def _auth_handler(self):
        should_warn = self.has_login_info()
        try:
            yield
        except authentication.FailedAuthentication as e:
            if should_warn:
                self.logger.warning(f"Invalid authentication details, please re-authenticate. {e}")
            self.logout()
        except authentication.UnavailableError:
            raise
        except Exception as e:
            self.logger.exception(e, True, f"Error when trying to refresh community login: {e}")

    def _ensure_email(self, email):
        if constants.USER_ACCOUNT_EMAIL and email != constants.USER_ACCOUNT_EMAIL:
            raise authentication.AuthenticationError("The given email doesn't match the expected user email.")

    def _ensure_community_url(self):
        if not self.can_authenticate():
            raise authentication.UnavailableError("Community url required")

    async def _on_account_updated(self):
        self.user_account.set_profile_raw_data(await self.supabase_client.get_user())

    def _reset_tokens(self):
        self.user_account.flush()

    @_bot_data_update
    async def update_trades(self, trades: list, exchange_name: str, reset: bool):
        """
        Updates authenticated account trades
        """
        if reset:
            await self.supabase_client.reset_trades(self.user_account.bot_id)
        trades_to_upload = trades if len(trades) <= self.MAX_UPLOADED_TRADES_COUNT else (
            sorted(
                trades,
                key=lambda x: x[trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value],
                reverse=True
            )[:self.MAX_UPLOADED_TRADES_COUNT]
        )
        if formatted_trades := formatters.format_trades(trades_to_upload, exchange_name, self.user_account.bot_id):
            await self.supabase_client.upsert_trades(formatted_trades)

    @_bot_data_update
    async def update_orders(self, orders: list, exchange_name: str):
        """
        Updates authenticated account orders
        """
        formatted_orders = formatters.format_orders(orders, exchange_name)
        await self.supabase_client.update_bot_orders(self.user_account.bot_id, formatted_orders)

    @_bot_data_update
    async def update_portfolio(self, current_value: dict, initial_value: dict, profitability: float,
                               unit: str, content: dict, history: dict, price_by_asset: dict,
                               reset: bool):
        """
        Updates authenticated account portfolio
        """
        try:
            formatted_portfolio = formatters.format_portfolio(
                current_value, initial_value, profitability, unit, content, price_by_asset, self.user_account.bot_id
            )
            if reset or self.user_account.get_selected_bot_current_portfolio_id() is None:
                await self.supabase_client.switch_portfolio(formatted_portfolio)
                await self.refresh_selected_bot()

            formatted_portfolio[backend_enums.PortfolioKeys.ID.value] = \
                self.user_account.get_selected_bot_current_portfolio_id()
            await self.supabase_client.update_portfolio(formatted_portfolio)
            if formatted_histories := formatters.format_portfolio_history(
                history, unit, self.user_account.get_selected_bot_current_portfolio_id()
            ):
                await self.supabase_client.upsert_portfolio_history(formatted_histories)
        except KeyError as err:
            self.logger.debug(f"Error when updating community portfolio {err} (missing reference market value)")

    @_bot_data_update
    async def update_bot_config_and_stats(self, profitability):
        formatted_portfolio = formatters.format_portfolio_with_profitability(profitability)
        if self.user_account.get_selected_bot_current_portfolio_id() is None:
            await self.refresh_selected_bot()
        if self.user_account.get_selected_bot_current_portfolio_id() is None:
            self.logger.debug(
                f"Skipping portfolio update: current bot {self.user_account.bot_id} has no current portfolio_id"
            )
        else:
            formatted_portfolio[backend_enums.PortfolioKeys.ID.value] = \
                self.user_account.get_selected_bot_current_portfolio_id()
            await self.supabase_client.update_portfolio(formatted_portfolio)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import logging
import math
import copy

import octobot_commons.data_util as data_util
import octobot_commons.tentacles_management as tentacles_management
import octobot_commons.logging as common_logging
import octobot_commons.constants as commons_constants
import octobot_commons.errors as commons_errors

import octobot.constants as constants
import octobot.strategy_optimizer as strategy_optimizer

import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_tentacles_manager.constants as tentacles_manager_constants

import octobot_evaluators.constants as evaluator_constants
import octobot_evaluators.evaluators as evaluators

import octobot_trading.modes as trading_modes

CONFIG = 0
RANK = 1
TRADES = 0
TRADES_IN_RESULT = 2


class StrategyOptimizer:
    """
    StrategyOptimizer is a tool that performs backtesting with different configurations
    """

    def __init__(self, config, tentacles_setup_config, strategy_name):
        # Lazy import of tentacles to let tentacles manager handle imports
        import tentacles.Evaluator as tentacles_Evaluator
        self.is_properly_initialized = False
        self.logger = common_logging.get_logger(self.get_name())
        self.config = config
        self.tentacles_setup_config = copy.deepcopy(tentacles_setup_config)
        self.trading_mode = trading_modes.get_activated_trading_mode(tentacles_setup_config)
        if self.trading_mode.get_name() not in self._get_compatible_trading_modes():
            raise commons_errors.ConfigTradingError(
                f"{self.trading_mode.get_name()} is not supported trading mode for the {self.get_name()}"
            )
        self.strategy_class = tentacles_management.get_class_from_string(
            strategy_name, evaluators.StrategyEvaluator,
            tentacles_Evaluator.Strategies, tentacles_management.evaluator_parent_inspection)
        self.run_results = []
        self.sorted_results_by_time_frame = {}
        self.sorted_results_through_all_time_frame = []
        self.all_time_frames = []
        self.all_TAs = []
        self.risks = []
        self.current_test_suite = None
        self.errors = set()

        self.is_computing = False
        self.is_finished = False
        self.run_id = 0
        self.total_nb_runs = 0

        self.keep_running = True

        if not self.strategy_class:
            self.logger.error(f"Impossible to find a strategy matching class name: {strategy_name} in installed "
                              f"strategies. Please make sure to enter the name of the class, "
                              f"ex: FullMixedStrategiesEvaluator")
        else:
            self.is_properly_initialized = True

    def find_optimal_configuration(self, TAs=None, time_frames=None, risks=None):
        if not self.is_computing:

            # set is_computing to True to prevent any simultaneous start
            self.is_computing = True

            self.errors = set()
            self.run_results = []
            self.sorted_results_by_time_frame = {}
            self.sorted_results_through_all_time_frame = []

            previous_log_level = common_logging.get_global_logger_level()
            previous_log_level_per_handler = common_logging.get_logger_level_per_handler()

            try:
                self.all_TAs = self._get_all_TA() if TAs is None else TAs
                nb_TAs = len(self.all_TAs)

                self.all_time_frames = self.strategy_class.get_required_time_frames(self.config,
                                                                                    self.tentacles_setup_config) \
                    if time_frames is None else time_frames
                nb_TFs = len(self.all_time_frames)

                self.risks = [1] if risks is None else risks

                self.logger.info(f"Trying to find an optimized configuration for {self.strategy_class.get_name()} "
                                 f"strategy using {self.trading_mode.get_name()} trading mode, {self.all_TAs} "
                                 f"technical evaluator(s), {self.all_time_frames} time frames and {self.risks} "
                                 f"risk(s).")

                self.total_nb_runs = int(len(self.risks) * ((math.pow(2, nb_TFs) - 1) * (math.pow(2, nb_TAs) - 1)))

                self.logger.info("Setting logging level to logging.ERROR to limit messages.")
                common_logging.set_global_logger_level(logging.ERROR)

                self.run_id = 1
                self._iterate_on_configs(nb_TAs, nb_TFs)
                self._find_optimal_configuration_using_results()
            finally:
                self.current_test_suite = None
                common_logging.set_global_logger_level(
                    previous_log_level,
                    handler_levels=previous_log_level_per_handler,
                )
                self.is_computing = False
                self.is_finished = True
                self.logger.info(f"{self.get_name()} finished computation.")
                self.logger.info("Logging level restored.")
        else:
            raise RuntimeError(f"{self.get_name()} is already computing: processed "
                               f"{self.run_id}/{self.total_nb_runs} processed")

    def _iterate_on_configs(self, nb_TAs, nb_TFs):
        # test with several risks
        for risk in self.risks:
            self.config[commons_constants.CONFIG_TRADING][commons_constants.CONFIG_TRADER_RISK] = risk
            eval_conf_history = []
            # test with several evaluators
            for evaluator_conf_iteration in range(nb_TAs):
                current_forced_evaluator = self.all_TAs[evaluator_conf_iteration]
                # test with 1-n evaluators at a time
                for nb_evaluators in range(1, nb_TAs + 1):
                    # test different configurations
                    for _ in range(nb_TAs):
                        activated_evaluators = self._get_activated_element(self.all_TAs,
                                                                           current_forced_evaluator,
                                                                           nb_evaluators, eval_conf_history,
                                                                           self.strategy_class.get_name(),
                                                                           True)
                        if activated_evaluators is not None:
                            self._adapt_tentacles_config(activated_evaluators)
                            time_frames_conf_history = []
                            # test different time frames
                            for time_frame_conf_iteration in range(nb_TFs):
                                current_forced_time_frame = self.all_time_frames[time_frame_conf_iteration]
                                # test with 1-n time frames at a time
                                for nb_time_frames in range(1, nb_TFs + 1):
                                    # test different configurations
                                    for _ in range(nb_TFs):
                                        if not self.keep_running:
                                            return
                                        self._run_on_config(risk, current_forced_time_frame, nb_time_frames,
                                                            time_frames_conf_history, activated_evaluators)

    def _run_on_config(self, risk, current_forced_time_frame, nb_time_frames,
                       time_frames_conf_history, activated_evaluators):
        activated_time_frames = self._get_activated_element(self.all_time_frames,
                                                            current_forced_time_frame,
                                                            nb_time_frames,
                                                            time_frames_conf_history)
        if activated_time_frames is not None:
            self.config[evaluator_constants.CONFIG_FORCED_TIME_FRAME] = activated_time_frames
            print(f"{self.run_id}/{self.total_nb_runs} Run with: evaluators: {activated_evaluators}, "
                  f"time frames :{activated_time_frames}, risk: {risk}")
            self._run_test_suite(self.config, activated_evaluators)
            print(f" => Result: {self.run_results[-1].get_result_string(False)}")
            self.run_id += 1

    def _run_test_suite(self, config, evaluators):
        self.current_test_suite = strategy_optimizer.StrategyTestSuite()
        self.current_test_suite.evaluators = list(evaluators)
        self.current_test_suite.initialize_with_strategy(self.strategy_class,
                                                         self.tentacles_setup_config,
                                                         copy.deepcopy(config))
        no_error = asyncio.run(self.current_test_suite.run_test_suite(self.current_test_suite),
                               debug=constants.OPTIMIZER_FORCE_ASYNCIO_DEBUG_OPTION)
        if not no_error:
            self.errors = self.errors.union(set([str(e) for e in self.current_test_suite.exceptions]))
        run_result = self.current_test_suite.get_test_suite_result()
        self.run_results.append(run_result)

    def _adapt_tentacles_config(self, activated_evaluators):
        # Lazy import of tentacles to let tentacles manager handle imports
        import tentacles.Evaluator as tentacles_Evaluator
        to_update_config = {}
        tentacles_activation = tentacles_manager_api.get_tentacles_activation(self.tentacles_setup_config)
        for tentacle_class_name in tentacles_activation[tentacles_manager_constants.TENTACLES_EVALUATOR_PATH]:
            if tentacle_class_name in activated_evaluators:
                to_update_config[tentacle_class_name] = True
            elif tentacles_management.get_class_from_string(tentacle_class_name, evaluators.StrategyEvaluator,
                                                            tentacles_Evaluator.Strategies,
                                                            tentacles_management.evaluator_parent_inspection) is None:
                to_update_config[tentacle_class_name] = False
        tentacles_manager_api.update_activation_configuration(self.tentacles_setup_config, to_update_config, False)

    def _find_optimal_configuration_using_results(self):
        for time_frame in self.all_time_frames:
            time_frame_sorted_results = self._get_sorted_results(self.run_results, time_frame)
            self.sorted_results_by_time_frame[time_frame.value] = time_frame_sorted_results

        results_through_all_time_frame = {}
        for results in self.sorted_results_by_time_frame.values():
            for rank, result in enumerate(results):
                result_summary = result.get_config_summary()
                if result_summary not in results_through_all_time_frame:
                    results_through_all_time_frame[result_summary] = [[], 0]
                results_through_all_time_frame[result_summary][RANK] += rank
                results_through_all_time_frame[result_summary][TRADES] += result.trades_counts

        result_list = [(result, trades_and_rank[RANK], data_util.mean(trades_and_rank[TRADES]))
                       for result, trades_and_rank in results_through_all_time_frame.items()]

        # inner function to replace uncythonizable lambda expression
        def _sorting_key(elem):
            return elem[RANK]

        self.sorted_results_through_all_time_frame = sorted(result_list, key=_sorting_key)

    def print_report(self):
        self.logger.info("Full execution sorted results: Minimum time frames are defining the range of the run "
                         "since it finishes at the end of the first data, aka minimum time frame. Therefore all "
                         "different time frames are different price actions and can't be compared independently.")
        for time_frame, results in self.sorted_results_by_time_frame.items():
            self.logger.info(f" *** {time_frame} minimum time frame ranking*** ")
            for rank, result in enumerate(results):
                self.logger.info(f"{rank}: {result.get_result_string()}")
        self.logger.info(f" *** Top rankings per time frame *** ")
        for time_frame, results in self.sorted_results_by_time_frame.items():
            for i in range(0, min(len(results), 5)):
                self.logger.info(f"{time_frame}: {results[i].get_result_string(False)}")
        self.logger.info(f" *** Top rankings through all time frames *** ")
        for rank, result in enumerate(self.sorted_results_through_all_time_frame[0:25]):
            self.logger.info(f"{rank}: {result[CONFIG].get_result_string()} (time frame rank sum: {result[RANK]}) "
                             f"average trades count: {result[TRADES_IN_RESULT]:f}")
        self.logger.info(f" *** Overall best configuration for {self.strategy_class.get_name()} using "
                         f"{self.trading_mode.get_name()} trading mode *** ")
        best_result = self.sorted_results_through_all_time_frame[0]
        self.logger.info(f"{best_result[CONFIG].get_result_string()} "
                         f"average trades count: {best_result[TRADES_IN_RESULT]:f}")

    async def get_overall_progress(self):
        return int((self.run_id - 1) / self.total_nb_runs * 100) if self.total_nb_runs else 0, 0

    async def is_in_progress(self):
        return self.get_overall_progress() != 100

    def cancel(self):
        self.keep_running = False

    def get_current_test_suite_progress(self):
        return self.current_test_suite.current_progress if self.current_test_suite else 0

    def get_report(self):
        # index, evaluators, risk, score, trades
        if self.sorted_results_through_all_time_frame:
            results = [strategy_optimizer.TestSuiteResult.convert_result_into_dict(rank, result[CONFIG].evaluators, "",
                                                                                   result[CONFIG].risk, result[RANK],
                                                                                   round(result[TRADES_IN_RESULT], 5))
                       for rank, result in enumerate(self.sorted_results_through_all_time_frame[0:100])]
        else:
            results = []
        return results

    def get_errors_description(self):
        if self.errors:
            return f"{', '.join(self.errors)[0:350]} ..."
        else:
            return None

    @classmethod
    def get_name(cls):
        return cls.__name__

    @staticmethod
    def _get_activated_element(all_elements, current_forced_element, nb_elements_to_consider,
                               elem_conf_history, default_element=None, dict_shaped=False):
        eval_conf = {current_forced_element: True}
        additional_elements_count = 0
        if default_element is not None:
            eval_conf[default_element] = True
            additional_elements_count = 1
        if nb_elements_to_consider > 1:
            i = 0
            while i < len(all_elements) and \
                    len(eval_conf) < nb_elements_to_consider + additional_elements_count:
                current_elem = all_elements[i]
                if current_elem not in eval_conf:
                    eval_conf[current_elem] = True
                    if len(eval_conf) == nb_elements_to_consider + additional_elements_count and \
                            ((eval_conf if dict_shaped else sorted([key.value for key in eval_conf]))
                             in elem_conf_history):
                        eval_conf.pop(current_elem)
                i += 1
        if len(eval_conf) == nb_elements_to_consider + additional_elements_count:
            to_use_conf = eval_conf
            if not dict_shaped:
                to_use_conf = sorted([key.value for key in eval_conf])
            if to_use_conf not in elem_conf_history:
                elem_conf_history.append(to_use_conf)
                return to_use_conf
        return None

    @staticmethod
    def _get_filtered_results(results, time_frame=None):
        return [result for result in results if time_frame is None or result.min_time_frame == time_frame]

    @staticmethod
    def _get_sorted_results(results, time_frame=None):
        return sorted(StrategyOptimizer._get_filtered_results(results, time_frame),
                      key=lambda result: result.get_average_score(), reverse=True)

    @staticmethod
    def _is_relevant_evaluation_config(evaluator):
        # Lazy import of tentacles to let tentacles manager handle imports
        import tentacles.Evaluator as tentacles_Evaluator
        return tentacles_management.get_class_from_string(
            evaluator, evaluators.TAEvaluator, tentacles_Evaluator.TA,
            tentacles_management.evaluator_parent_inspection) is not None

    def _get_all_TA(self):
        return [evaluator
                for evaluator, activated
                in tentacles_manager_api.get_tentacles_activation(self.tentacles_setup_config)[
                    tentacles_manager_constants.TENTACLES_EVALUATOR_PATH].items()
                if activated and StrategyOptimizer._is_relevant_evaluation_config(evaluator)]

    def _get_compatible_trading_modes(self):
        # Lazy import of tentacles to let tentacles manager handle imports
        import tentacles.Trading.Mode as modes
        return [modes.DailyTradingMode.get_name()]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from octobot.community.supabase_backend import postgres_functions
from octobot.community.supabase_backend.postgres_functions import (
    PostgresFunctions,
)
from octobot.community.supabase_backend import configuration_storage
from octobot.community.supabase_backend.configuration_storage import (
    SyncConfigurationStorage,
    ASyncConfigurationStorage,
)
from octobot.community.supabase_backend import supabase_client
from octobot.community.supabase_backend.supabase_client import (
    AuthenticatedAsyncSupabaseClient,
)
from octobot.community.supabase_backend import community_supabase_client
from octobot.community.supabase_backend.community_supabase_client import (
    CommunitySupabaseClient,
)

__all__ = [
    "PostgresFunctions",
    "SyncConfigurationStorage",
    "ASyncConfigurationStorage",
    "AuthenticatedAsyncSupabaseClient",
    "CommunitySupabaseClient",
]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import pytest
import time

import octobot.community as community


ERROR_TITLE = "An error happened"
ERROR_METRICS_ID = "1254xyz"
ERROR_TIME = time.time()
UPLOADER_URL = "http://upload_url"


@pytest.fixture
def basic_error():
    return community.Error(
        None,
        ERROR_TITLE,
        ERROR_TIME,
        ERROR_METRICS_ID
    )


@pytest.fixture
def exception_error():
    # generated exception with traceback
    return community.Error(
        _get_exception(),
        ERROR_TITLE,
        ERROR_TIME,
        ERROR_METRICS_ID
    )


@pytest.fixture
def error_uploader():
    return community.ErrorsUploader(UPLOADER_URL)


def _get_exception():
    def fake3():
        1/0

    def fake2():
        fake3()

    def fake_func():
        fake2()
    try:
        fake_func()
    except ZeroDivisionError as err:
        return err



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import copy

import octobot_commons.data_util as data_util
import octobot_commons.time_frame_manager as time_frame_manager


class TestSuiteResult:

    BOT_PROFITABILITY = 0
    MARKET_PROFITABILITY = 1

    INDEX = "id"
    EVALUATORS = "evaluators"
    TIME_FRAMES = "time_frames"
    RISK = "risk"
    SCORE = "score"
    AVERAGE_TRADES = "average_trades"

    def __init__(self, run_profitabilities, trades_counts, risk, time_frames, evaluators, strategy):
        self.run_profitabilities = run_profitabilities
        self.trades_counts = trades_counts
        self.risk = risk
        self.time_frames = time_frames
        self.min_time_frame = time_frame_manager.find_min_time_frame(self.time_frames)
        self.evaluators = evaluators
        self.strategy = strategy

    def get_average_score(self):
        bot_profitabilities = [
            float(profitability_result[self.BOT_PROFITABILITY] - profitability_result[self.MARKET_PROFITABILITY])
            for profitability_result in self.run_profitabilities]
        return data_util.mean(bot_profitabilities)

    def get_average_trades_count(self):
        return data_util.mean(self.trades_counts)

    def get_evaluators_without_strategy(self):
        evals = copy.copy(self.evaluators)
        evals.remove(self.strategy)
        return [eval_name for eval_name in evals]

    def get_config_summary(self):
        return TestSuiteResultSummary(self)

    def get_result_string(self, details=True):
        details_str = f" details: (profitabilities (bot, market):{self.run_profitabilities}, trades: " \
                      f"{self.trades_counts})" if details else ""
        return (f"{self.get_evaluators_without_strategy()} on {self.time_frames} at risk: {self.risk} "
                f"score: {self.get_average_score():f} (the higher the better) "
                f"average trades: {self.get_average_trades_count():f}{details_str}")

    def get_result_dict(self, index=0):
        return self.convert_result_into_dict(index, self.get_evaluators_without_strategy(), self.time_frames,
                                             self.risk, round(self.get_average_score(), 5),
                                             round(self.get_average_trades_count(), 5))

    @staticmethod
    def convert_result_into_dict(index, evaluators, time_frames, risk, score, trades):
        return {
            TestSuiteResult.INDEX: index,
            TestSuiteResult.EVALUATORS: evaluators,
            TestSuiteResult.TIME_FRAMES: time_frames,
            TestSuiteResult.RISK: risk,
            TestSuiteResult.SCORE: score,
            TestSuiteResult.AVERAGE_TRADES: trades,
        }


class TestSuiteResultSummary:
    def __init__(self, test_suite_result):
        self.evaluators = test_suite_result.get_evaluators_without_strategy()
        self.risk = test_suite_result.risk

    def get_result_string(self):
        return f"{self.evaluators} risk: {self.risk}"

    def __eq__(self, other):
        return self.evaluators == other.evaluators and self.risk == other.risk

    def __hash__(self):
        return abs(hash(f"{self.evaluators}{self.risk}"))



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import typing
import packaging.version

import octobot.community.identifiers_provider as identifiers_provider
import octobot.constants as constants


class CommunityTentaclesPackage:
    def __init__(self, name, description, url, activated, images, download_url, versions, last_version):
        self.name: str = name
        self.description: str = description
        self.url: str = url
        self.activated: bool = activated
        self.images: typing.List[str] = images
        self.download_url: str = download_url
        self.uninstalled: bool = not self.is_installed()
        self.versions: typing.List[str] = versions
        self.last_version: str = last_version

    @staticmethod
    def from_community_dict(data):
        data_attributes = data["attributes"]
        # todo update with new urls
        return CommunityTentaclesPackage(
            data_attributes.get("name"),
            data_attributes.get("description"),
            f"{identifiers_provider.IdentifiersProvider.COMMUNITY_URL}products/{data_attributes.get('product_slug')}",
            data_attributes.get("activated"),
            data["relationships"].get('images')['data'],
            f"{identifiers_provider.IdentifiersProvider.COMMUNITY_URL}{data_attributes.get('download_path')}",
            data_attributes.get("versions"),
            data_attributes.get("last_version")
        )

    def get_latest_compatible_version(self):
        current_bot_version = packaging.version.parse(constants.LONG_VERSION)
        if packaging.version.parse(self.last_version) <= current_bot_version:
            return self.last_version
        available_versions = sorted([packaging.version.parse(version) for version in self.versions], reverse=True)
        for version in available_versions:
            if version <= current_bot_version:
                return version
        return None

    def is_installed(self):
        #TODO tmp
        import random
        return random.choice((True, False))



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot.community.supabase_backend.enums as backend_enums
import octobot.community.supabase_backend as supabase_backend
import octobot_commons.constants as commons_constants
import octobot_trading.enums as trading_enums
import octobot_trading.constants as trading_constants
import octobot_trading.personal_data as trading_personal_data


def format_trades(trades: list, exchange_name: str, bot_id: str) -> list:
    return [
        _format_trade(trade, exchange_name, bot_id)
        for trade in trades
    ]


def _format_trade(trade: dict, exchange_name: str, bot_id: str):
    metadata = {
        trading_enums.ExchangeConstantsOrderColumns.ENTRIES.value:
            trade[trading_enums.ExchangeConstantsOrderColumns.ENTRIES.value]
    }
    return {
            backend_enums.TradeKeys.BOT_ID.value: bot_id,
            backend_enums.TradeKeys.TRADE_ID.value: trade[trading_enums.ExchangeConstantsOrderColumns.EXCHANGE_ID.value]
            or trade[trading_enums.ExchangeConstantsOrderColumns.ID.value],
            backend_enums.TradeKeys.TIME.value: supabase_backend.CommunitySupabaseClient.get_formatted_time(
                trade[trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value]
            ),
            backend_enums.TradeKeys.EXCHANGE.value: exchange_name,
            backend_enums.TradeKeys.PRICE.value: float(trade[trading_enums.ExchangeConstantsOrderColumns.PRICE.value]),
            backend_enums.TradeKeys.QUANTITY.value:
                float(trade[trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value]),
            backend_enums.TradeKeys.SYMBOL.value: trade[trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value],
            backend_enums.TradeKeys.VOLUME.value: float(
                trade.get(trading_enums.ExchangeConstantsOrderColumns.VOLUME.value, 0)
            ),
            backend_enums.TradeKeys.TYPE.value: _get_order_type(trade),
            backend_enums.TradeKeys.METADATA.value: metadata
        }


def format_orders(orders: list, exchange_name: str) -> list:
    return [
        {
            backend_enums.OrderKeys.EXCHANGE.value: exchange_name,
            backend_enums.OrderKeys.SYMBOL.value: storage_order[trading_constants.STORAGE_ORIGIN_VALUE][
                trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value],
            backend_enums.OrderKeys.PRICE.value: storage_order[trading_constants.STORAGE_ORIGIN_VALUE][
                trading_enums.ExchangeConstantsOrderColumns.PRICE.value],
            backend_enums.OrderKeys.TIME.value: storage_order[trading_constants.STORAGE_ORIGIN_VALUE][
                trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value],
            backend_enums.OrderKeys.TYPE.value: _get_order_type(
                storage_order[trading_constants.STORAGE_ORIGIN_VALUE]
            ),
            backend_enums.OrderKeys.QUANTITY.value: storage_order[trading_constants.STORAGE_ORIGIN_VALUE][
                trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value],
        }
        for storage_order in orders
    ]


def _get_order_type(order_or_trade):
    order_type = order_or_trade[trading_enums.ExchangeConstantsOrderColumns.TYPE.value]
    try:
        return trading_personal_data.parse_order_type(order_or_trade)[1].value
    except Exception:
        # use default trade_type
        return order_type


def format_portfolio(
    current_value: dict, initial_value: dict, profitability: float,
    unit: str, content: dict, price_by_asset: dict,
    bot_id: str
) -> dict:
    ref_market_current_value = current_value[unit]
    ref_market_initial_value = initial_value[unit]
    formatted_content = [
        {
            backend_enums.PortfolioAssetKeys.ASSET.value: key,
            backend_enums.PortfolioAssetKeys.QUANTITY.value: float(quantity[commons_constants.PORTFOLIO_TOTAL]),
            backend_enums.PortfolioAssetKeys.VALUE.value:
                float(quantity[commons_constants.PORTFOLIO_TOTAL]) * float(price_by_asset.get(key, 0)),
        }
        for key, quantity in content.items()
    ]

    return {
        backend_enums.PortfolioKeys.CONTENT.value: formatted_content,
        backend_enums.PortfolioKeys.CURRENT_VALUE.value: ref_market_current_value,
        backend_enums.PortfolioKeys.INITIAL_VALUE.value: ref_market_initial_value,
        backend_enums.PortfolioKeys.PROFITABILITY.value: float(profitability),
        backend_enums.PortfolioKeys.UNIT.value: unit,
        backend_enums.PortfolioKeys.BOT_ID.value: bot_id,
    }


def format_portfolio_with_profitability(profitability) -> dict:
    return {
        backend_enums.PortfolioKeys.PROFITABILITY.value: float(profitability)
    }


def format_portfolio_history(history: dict, unit: str, portfolio_id: str) -> list:
    try:
        return [
            {
                backend_enums.PortfolioHistoryKeys.TIME.value:
                    supabase_backend.CommunitySupabaseClient.get_formatted_time(timestamp),
                backend_enums.PortfolioHistoryKeys.PORTFOLIO_ID.value: portfolio_id,
                backend_enums.PortfolioHistoryKeys.VALUE.value: float(value[unit])
            }
            for timestamp, value in history.items()
            if unit in value and value[unit]    # skip missing a 0 values
        ]
    except KeyError:
        return []



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import pytest
import pytest_asyncio
import mock
import uuid
import json
import zlib
import gmqtt

import octobot.community as community
import octobot.constants as constants
import octobot_commons.enums as commons_enums

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

FEED_URL = "x.y.z"
TOKEN = "acb1"
NAME = "name_a"


def _build_message(value, identifier):
    return {
        commons_enums.CommunityFeedAttrs.CHANNEL_TYPE.value: commons_enums.CommunityChannelTypes.SIGNAL.value,
        commons_enums.CommunityFeedAttrs.VERSION.value: constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION,
        commons_enums.CommunityFeedAttrs.VALUE.value: value,
        commons_enums.CommunityFeedAttrs.ID.value: identifier,
    }


def _zipped_message(value):
    return zlib.compress(
        json.dumps(value).encode()
    )


@pytest_asyncio.fixture
async def authenticator():
    community.IdentifiersProvider.use_production()
    auth = community.CommunityAuthentication(None, None)
    auth._auth_token = TOKEN
    auth.refresh_token = TOKEN
    auth._expire_at = 11
    return auth


@pytest_asyncio.fixture
async def connected_community_feed(authenticator):
    feed = None
    try:
        feed = community.CommunityMQTTFeed(FEED_URL, authenticator)
        feed.INIT_TIMEOUT = 1
        with mock.patch.object(authenticator.user_account, "get_selected_bot_device_uuid", mock.Mock(return_value=TOKEN)) \
                as get_selected_bot_device_uuid_mock, \
             mock.patch.object(authenticator.user_account, "get_selected_bot_device_name", mock.Mock(return_value=NAME)
                               ) as _get_selected_bot_device_uuid_mock, \
             mock.patch.object(feed, "_subscribe", mock.Mock()) as _subscribe_mock, \
             mock.patch.object(gmqtt.Client, "connect", mock.AsyncMock()) as _connect_mock:
            await feed.register_feed_callback(commons_enums.CommunityChannelTypes.SIGNAL, mock.AsyncMock())
            _subscribe_mock.assert_called_once_with((f"{commons_enums.CommunityChannelTypes.SIGNAL.value}/None", ))
            await feed.start()
            get_selected_bot_device_uuid_mock.assert_called_once()
            _connect_mock.assert_called_once_with(FEED_URL, feed.mqtt_broker_port, version=feed.MQTT_VERSION)
            yield feed
    finally:
        if feed is not None:
            await feed.stop()
            if feed._mqtt_client is not None and not feed._mqtt_client._resend_task.done():
                feed._mqtt_client._resend_task.cancel()


async def test_start_and_connect(connected_community_feed):
    # connected_community_feed already called _connect, check that everything has been set
    assert isinstance(connected_community_feed._mqtt_client, gmqtt.Client)
    assert connected_community_feed._mqtt_client._config["reconnect_retries"] == \
           connected_community_feed.DISABLE_RECONNECT_VALUE
    assert connected_community_feed._mqtt_client.on_connect == connected_community_feed._on_connect
    assert connected_community_feed._mqtt_client.on_message == connected_community_feed._on_message
    assert connected_community_feed._mqtt_client.on_disconnect == connected_community_feed._on_disconnect
    assert connected_community_feed._mqtt_client.on_subscribe == connected_community_feed._on_subscribe
    assert connected_community_feed._mqtt_client._username == TOKEN.encode()


async def test_stop(connected_community_feed):
    connected_community_feed._reconnect_task = None
    with mock.patch.object(connected_community_feed, "_stop_mqtt_client", mock.AsyncMock()) as _stop_mqtt_client_mock:
        await connected_community_feed.stop()
        _stop_mqtt_client_mock.assert_called_once()

        _stop_mqtt_client_mock.reset_mock()
        connected_community_feed._reconnect_task = mock.Mock(done=mock.Mock(return_value=True), cancel=mock.Mock())
        await connected_community_feed.stop()
        _stop_mqtt_client_mock.assert_called_once()
        connected_community_feed._reconnect_task.done.assert_called_once()
        connected_community_feed._reconnect_task.cancel.assert_not_called()

        _stop_mqtt_client_mock.reset_mock()
        connected_community_feed._reconnect_task = mock.Mock(done=mock.Mock(return_value=False), cancel=mock.Mock())
        await connected_community_feed.stop()
        _stop_mqtt_client_mock.assert_called_once()
        connected_community_feed._reconnect_task.done.assert_called_once()
        connected_community_feed._reconnect_task.cancel.assert_called_once()


async def test_register_feed_callback(connected_community_feed):
    # TODO
    pass


async def test_on_message(connected_community_feed):
    client = mock.Mock(client_id="1")
    topic = "topic"
    connected_community_feed.feed_callbacks["topic"] = [mock.AsyncMock(), mock.AsyncMock()]

    message = _build_message("hello", "1")
    # from topic
    await connected_community_feed._on_message(client, "other_topic", _zipped_message(message), 1, {})
    assert all(cb.assert_not_called() is None for cb in connected_community_feed.feed_callbacks["topic"])

    message = _build_message("hello", "2")
    # call callbacks
    await connected_community_feed._on_message(client, topic, _zipped_message(message), 1, {})
    assert all(cb.assert_called_once_with(message) is None for cb in connected_community_feed.feed_callbacks["topic"])

    # already processed message
    connected_community_feed.feed_callbacks["topic"][0].reset_mock()
    connected_community_feed.feed_callbacks["topic"][1].reset_mock()
    await connected_community_feed._on_message(client, topic, _zipped_message(message), 1, {})
    assert all(cb.assert_not_called() is None for cb in connected_community_feed.feed_callbacks["topic"])


async def test_send(connected_community_feed):
    with mock.patch.object(connected_community_feed._mqtt_client, "publish", mock.Mock()) as publish_mock, \
         mock.patch.object(uuid, "uuid4", mock.Mock(return_value="uuid41")) as uuid4_mock:
        await connected_community_feed.send("hello", commons_enums.CommunityChannelTypes.SIGNAL, "x")
        publish_mock.assert_called_once_with(
            f"{commons_enums.CommunityChannelTypes.SIGNAL.value}/x",
            zlib.compress(json.dumps({
                commons_enums.CommunityFeedAttrs.CHANNEL_TYPE.value: commons_enums.CommunityChannelTypes.SIGNAL.value,
                commons_enums.CommunityFeedAttrs.VERSION.value: constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION,
                commons_enums.CommunityFeedAttrs.VALUE.value: "hello",
                commons_enums.CommunityFeedAttrs.ID.value: "uuid41",  # assign unique id to each message
            }).encode()),
            qos=connected_community_feed.default_QOS
        )
        uuid4_mock.assert_called_once()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import uuid
import json
import realtime
import packaging.version as packaging_version

import octobot_commons.enums as commons_enums
import octobot_commons.authentication as authentication
import octobot_commons.errors as commons_errors
import octobot.community.supabase_backend.enums as enums
import octobot.community.feeds.abstract_feed as abstract_feed
import octobot.constants as constants


class CommunitySupabaseFeed(abstract_feed.AbstractFeed):
    SCHEMA = "public"
    SIGNALS_TABLE = "signals"
    INSERT_EVENT = "INSERT"
    MAX_MESSAGE_ID_CACHE_SIZE = 100

    def __init__(self, feed_url, authenticator):
        super().__init__(feed_url, authenticator)
        self._realtime_client = authenticator.supabase_client.realtime
        self._processed_messages = []

    def _ensure_supported(self, parsed_message):
        if packaging_version.Version(parsed_message[commons_enums.CommunityFeedAttrs.VERSION.value]) \
                < packaging_version.Version(constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION):
            raise commons_errors.UnsupportedError(
                f"Minimum version: {constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION}"
            )

    def _should_process(self, parsed_message):
        if parsed_message[commons_enums.CommunityFeedAttrs.ID.value] in self._processed_messages:
            self.logger.debug(f"Ignored already processed message with id: "
                              f"{parsed_message[commons_enums.CommunityFeedAttrs.ID.value]}")
            return False
        self._processed_messages.append(parsed_message[commons_enums.CommunityFeedAttrs.ID.value])
        if len(self._processed_messages) > self.MAX_MESSAGE_ID_CACHE_SIZE:
            self._processed_messages = [
                message_id
                for message_id in self._processed_messages[self.MAX_MESSAGE_ID_CACHE_SIZE // 2:]
            ]
        return True

    def _get_callbacks(self, table, identifier):
        for callback in self.feed_callbacks.get(table, {}).get(identifier, tuple()):
            yield callback

    async def _process_message(self, table: str, message: dict):
        try:
            content = message["new"]
            parsed_message = content["signal"]
            self._ensure_supported(parsed_message)
            if self._should_process(parsed_message):
                self.update_last_message_time()
                product_id = content[enums.SignalKeys.PRODUCT_ID.value]
                for callback in self._get_callbacks(table, product_id):
                    await callback(parsed_message)
        except commons_errors.UnsupportedError as err:
            self.logger.error(f"Unsupported message: {err}")
        except Exception as err:
            self.logger.exception(err, True, f"Unexpected error when processing message: {err}")

    async def start(self):
        # handled in supabase client directly, just ensure no subscriptions are pending
        for table in self.feed_callbacks:
            await self._subscribe_to_table_if_necessary(table)

    async def stop(self):
        # handled in supabase client directly
        pass

    async def register_feed_callback(self, channel_type: commons_enums.CommunityChannelTypes, callback, identifier=None):
        self.is_signal_receiver = True
        table = self._get_table(channel_type)
        try:
            self.feed_callbacks[table][identifier].append(callback)
        except KeyError:
            if table not in self.feed_callbacks:
                self.feed_callbacks[table] = {}
            if identifier not in self.feed_callbacks[table]:
                self.feed_callbacks[table][identifier] = [callback]
        await self._subscribe_to_table_if_necessary(table)

    async def _subscribe_to_table_if_necessary(self, table):
        if table not in self.authenticator.supabase_client.get_subscribed_channel_tables():
            if self.authenticator.is_logged_in():
                await self._subscribe_to_table(table)
            else:
                raise authentication.AuthenticationRequired("You need to be authenticated to be able to "
                                                            "connect to signals")

    def _get_table(self, channel_type: commons_enums.CommunityChannelTypes):
        if channel_type is commons_enums.CommunityChannelTypes.SIGNAL:
            return self.SIGNALS_TABLE
        raise NotImplementedError(channel_type)

    async def _subscribe_to_table(self, table):
        channel = await self._realtime_client.channel(self.SCHEMA, table)
        await channel.on(self.INSERT_EVENT, self._message_callback).subscribe(self._subscribed_callback)

    async def _message_callback(self, message: dict):
        self.logger.debug(f"Received message: {message['new']}")
        await self._process_message(message["table"], message)

    async def _subscribed_callback(self, state: str, message: realtime.Message):
        self.logger.debug(f"Subscribe to {self._get_table_from_message(message)}: {state}")

    def _get_table_from_message(self, message: realtime.Message):
        return message.topic.split(":")[-1]

    async def send(self, message, channel_type, identifier, **kwargs):
        self.is_signal_emitter = True
        if not self.authenticator.is_logged_in():
            self.logger.warning(f"Can't send {channel_type.name}, invalid feed authentication.")
            return
        self.logger.info(f"Sending signal on identifier: {identifier}, message: {message}")
        await self.authenticator.supabase_client.send_signal(
            self._get_table(channel_type),
            identifier,
            self._build_message(channel_type, message)
        )

    def _build_message(self, channel_type, message):
        if message:
            return json.dumps({
                commons_enums.CommunityFeedAttrs.CHANNEL_TYPE.value: channel_type.value,
                commons_enums.CommunityFeedAttrs.VERSION.value: constants.COMMUNITY_FEED_CURRENT_MINIMUM_VERSION,
                commons_enums.CommunityFeedAttrs.VALUE.value: message,
                commons_enums.CommunityFeedAttrs.ID.value: str(uuid.uuid4()),
                # assign unique id to each message
            })
        return {}

    def is_connected_to_remote_feed(self):
        return self.authenticator.supabase_client.is_realtime_connected()

    def is_connected(self):
        return self.authenticator.is_logged_in()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.


class CommunityDonation:
    def __init__(self, amount: str, currency: str, blockchain: str, transaction_id: str, address_to: str):
        self.amount = amount
        self.currency = currency
        self.blockchain = blockchain
        self.transaction_id = transaction_id
        self.address_to = address_to

    def __str__(self):
        return f"{self.amount} {self.currency} on {self.blockchain} ({self.transaction_id})"

    @staticmethod
    def from_community_dict(data):
        data_attributes = data.get("attributes", {})
        return CommunityDonation(
            data_attributes.get("amount"),
            data_attributes.get("currency"),
            data_attributes.get("blockchain"),
            data_attributes.get("transaction_id"),
            data_attributes.get("address_to")
        )



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio

import octobot_commons.constants as commons_constants
from octobot.api.backtesting import create_independent_backtesting, initialize_and_run_independent_backtesting, \
    check_independent_backtesting_remaining_objects, stop_independent_backtesting, join_independent_backtesting, \
    get_independent_backtesting_exchange_manager_ids
from octobot.backtesting.abstract_backtesting_test import DATA_FILES
from octobot_commons.asyncio_tools import ErrorContainer
from octobot_trading.api.exchange import get_exchange_manager_from_exchange_id
from octobot_trading.api.orders import get_open_orders
from octobot_trading.api.trades import get_trade_history


async def run_independent_backtestings_with_memory_check(config, tentacles_setup_config, backtesting_count=3):
    """
    Will raise an error of an object is not deleted by garbage collector at the end of the backtesting process
    :param config: the global config to use in backtesting
    :param tentacles_setup_config: the tentacles setup config to use (with the tentacles to be tested)
    :param backtesting_count: number of backtestings to run to ensure no side effects, default is 3
    :return:
    """
    backtesting = None
    try:
        config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]["USDT"] = 10000
        config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]["ETH"] = 20
        for _ in range(backtesting_count):
            error_container = ErrorContainer()
            asyncio.get_event_loop().set_exception_handler(error_container.exception_handler)
            # enabling loggers is slowing down backtesting but can display useful debug info
            # from octobot.logger import init_logger
            # init_logger()
            backtesting = await _run_backtesting(config, tentacles_setup_config)
            exchange_manager = get_exchange_manager_from_exchange_id(
                get_independent_backtesting_exchange_manager_ids(backtesting)[0])
            trades = get_trade_history(exchange_manager)
            open_orders = get_open_orders(exchange_manager)
            # ensure at least one order is either open or got filled
            assert trades + open_orders
            trades = open_orders = exchange_manager = None  # prevent memory leak
            await stop_independent_backtesting(backtesting, memory_check=True)
            await asyncio.wait_for(backtesting.post_backtesting_task, 5)
            asyncio.get_event_loop().call_soon(check_independent_backtesting_remaining_objects, backtesting)
            await asyncio.create_task(error_container.check())
    except Exception as e:
        if backtesting is not None:
            # do not get stuck in running backtesting
            await stop_independent_backtesting(backtesting, memory_check=False)
            await asyncio.wait_for(backtesting.post_backtesting_task, 5)
        raise e


async def _run_backtesting(config, tentacles_setup_config):
    backtesting = create_independent_backtesting(config, tentacles_setup_config, [DATA_FILES["ETH/USDT"]], "",
                                                 enable_storage=False)
    await initialize_and_run_independent_backtesting(backtesting, log_errors=False)
    await join_independent_backtesting(backtesting)
    return backtesting



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from octobot.strategy_optimizer import test_suite_result
from octobot.strategy_optimizer import strategy_optimizer
from octobot.strategy_optimizer import strategy_design_optimizer
from octobot.strategy_optimizer import strategy_test_suite

from octobot.strategy_optimizer.test_suite_result import (
    TestSuiteResult,
    TestSuiteResultSummary,
)
from octobot.strategy_optimizer.strategy_optimizer import (
    StrategyOptimizer,
)
from octobot.strategy_optimizer.fitness_parameter import (
    FitnessParameter,
)
from octobot.strategy_optimizer.optimizer_filter import (
    OptimizerFilter,
)
from octobot.strategy_optimizer.optimizer_settings import (
    OptimizerSettings,
)
from octobot.strategy_optimizer.scored_run_result import (
    ScoredRunResult,
)
from octobot.strategy_optimizer.optimizer_constraint import (
    OptimizerConstraint,
)
from octobot.strategy_optimizer.strategy_design_optimizer import (
    StrategyDesignOptimizer,
)
from octobot.strategy_optimizer.strategy_test_suite import (
    StrategyTestSuite,
)
from octobot.strategy_optimizer.strategy_design_optimizer_factory import (
    create_most_advanced_strategy_design_optimizer,
)

__all__ = [
    "TestSuiteResult",
    "TestSuiteResultSummary",
    "StrategyOptimizer",
    "FitnessParameter",
    "OptimizerFilter",
    "OptimizerSettings",
    "ScoredRunResult",
    "OptimizerConstraint",
    "StrategyDesignOptimizer",
    "StrategyTestSuite",
    "create_most_advanced_strategy_design_optimizer",
]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import math
import numpy

import octobot_commons.enums as common_enums
import octobot_commons.constants as commons_constants
import octobot_commons.databases as commons_databases
import octobot_commons.configuration as commons_configuration
import octobot_commons.time_frame_manager as time_frame_manager

import octobot_backtesting.api as backtesting_api

import octobot_trading.api as trading_api
import octobot_trading.enums as trading_enums
import octobot.backtesting as backtesting


async def clear_run_metadata(bot_id):
    run_db = commons_databases.RunDatabasesProvider.instance().get_run_db(bot_id)
    await run_db.delete(common_enums.DBTables.METADATA.value, None)


async def store_run_metadata(bot_id, exchange_managers, start_time, user_inputs=None, flush=False) -> dict:
    run_db = commons_databases.RunDatabasesProvider.instance().get_run_db(bot_id)
    user_inputs = user_inputs or await commons_configuration.get_user_inputs(run_db)
    run_dbs_identifier = commons_databases.RunDatabasesProvider.instance().get_run_databases_identifier(bot_id)
    metadata = await _get_trading_metadata(exchange_managers, start_time, user_inputs, run_dbs_identifier, False, None)
    await run_db.log(
        common_enums.DBTables.METADATA.value,
        metadata
    )
    if flush:
        await run_db.flush()
    return metadata


async def store_backtesting_run_metadata(exchange_managers, start_time, user_inputs, run_dbs_identifier, name) -> dict:
    run_metadata = await _get_trading_metadata(exchange_managers, start_time, user_inputs, run_dbs_identifier, True, name)
    # use local database as a lock is required
    async with commons_databases.DBWriter.database(
            run_dbs_identifier.get_backtesting_metadata_identifier(),
            with_lock=True) as writer:
        await writer.log(common_enums.DBTables.METADATA.value, run_metadata)
    return run_metadata


async def _get_trading_metadata(exchange_managers, run_start_time, user_inputs, run_dbs_identifier, is_backtesting, name) \
        -> dict:
    trading_mode = trading_api.get_trading_modes(exchange_managers[0])[0]
    multi_exchanges_data = await _get_multi_exchange_data(exchange_managers, is_backtesting)
    single_exchange_data = await _get_single_exchange_data(
        exchange_managers[0],
        trading_mode,
        run_start_time,
        user_inputs,
        run_dbs_identifier,
        is_backtesting,
        name
    )
    return {
        **single_exchange_data,
        **multi_exchanges_data,
        **(await trading_mode.get_additional_metadata(is_backtesting))
    }


async def _get_multi_exchange_data(exchange_managers, is_backtesting):
    symbols = list(set(
        symbol
        for exchange_manager in exchange_managers
        for symbol in trading_api.get_trading_pairs(exchange_manager)
    ))
    time_frames = list(set(
        tf.value
        for exchange_manager in exchange_managers
        for tf in trading_api.get_relevant_time_frames(exchange_manager)
    ))
    data_files = list(set(
        data_file
        for exchange_manager in exchange_managers
        for data_file in trading_api.get_backtesting_data_files(exchange_manager)
    )) if is_backtesting else []
    profitability = numpy.average(tuple(
        float(trading_api.get_profitability_stats(exchange_manager)[0])
        for exchange_manager in exchange_managers
    ))
    profitability_percent = numpy.average(tuple(
        float(trading_api.get_profitability_stats(exchange_manager)[1])
        for exchange_manager in exchange_managers
    ))
    markets_profitability = _get_markets_profitability(exchange_managers, time_frames) if is_backtesting else {}
    exchange_names = [
        trading_api.get_exchange_name(exchange_manager)
        for exchange_manager in exchange_managers
    ]
    future_contracts_by_exchange = _get_future_contracts_by_exchange(exchange_managers)
    trades = [
        trade
        for exchange_manager in exchange_managers
        for trade in trading_api.get_trade_history(exchange_manager, include_cancelled=False)
    ]
    entries = [
        trade
        for trade in trades
        if trade.status is trading_enums.OrderStatus.FILLED and trade.side is trading_enums.TradeOrderSide.BUY
    ]
    win_rate = round(numpy.average(tuple(
        float(trading_api.get_win_rate(exchange_manager) * 100)
        for exchange_manager in exchange_managers
    )), 3)
    wins = 0 if math.isnan(win_rate) else round(win_rate * len(entries) / 100)
    draw_down = round(numpy.average(tuple(
        float(trading_api.get_draw_down(exchange_manager))
        for exchange_manager in exchange_managers
    )), 3)
    r_sq_end_balance = await _get_coefficient_of_determination(exchange_managers, False)
    r_sq_max_balance = await _get_coefficient_of_determination(exchange_managers, True)
    duration = round(max(
        backtesting_api.get_backtesting_duration(exchange_manager.exchange.backtesting)
        for exchange_manager in exchange_managers
    ), 3) if is_backtesting else 0
    if trading_api.get_is_backtesting(exchange_managers[0]):
        start_time = min(
            backtesting_api.get_backtesting_starting_time(exchange_manager.exchange.backtesting)
            for exchange_manager in exchange_managers
        )
        end_time = max(
            backtesting_api.get_backtesting_ending_time(exchange_manager.exchange.backtesting)
            for exchange_manager in exchange_managers
        )
    else:
        start_time = exchange_managers[0].exchange.get_exchange_current_time()
        end_time = -1
    origin_portfolio, end_portfolio = _get_portfolio(exchange_managers)
    backtesting_only_metadata = {
        common_enums.BacktestingMetadata.DURATION.value: duration,
        common_enums.BacktestingMetadata.BACKTESTING_FILES.value: data_files,
    } if is_backtesting else {}
    return {
        **backtesting_only_metadata,
        **{
            common_enums.BacktestingMetadata.GAINS.value: round(float(profitability), 8),
            common_enums.BacktestingMetadata.PERCENT_GAINS.value: round(float(profitability_percent), 3),
            common_enums.BacktestingMetadata.MARKETS_PROFITABILITY.value:
                {c: f"{round(float(v) * 100, 2)}%" for c, v in markets_profitability.items()},
            common_enums.BacktestingMetadata.END_PORTFOLIO.value: str(end_portfolio),
            common_enums.BacktestingMetadata.START_PORTFOLIO.value: str(origin_portfolio),
            common_enums.BacktestingMetadata.WIN_RATE.value: win_rate,
            common_enums.BacktestingMetadata.DRAW_DOWN.value: draw_down or 0,
            common_enums.BacktestingMetadata.COEFFICIENT_OF_DETERMINATION_MAX_BALANCE.value: r_sq_max_balance or 0,
            common_enums.BacktestingMetadata.COEFFICIENT_OF_DETERMINATION_END_BALANCE.value: r_sq_end_balance or 0,
            common_enums.BacktestingMetadata.SYMBOLS.value: symbols,
            common_enums.BacktestingMetadata.TIME_FRAMES.value: time_frames,
            common_enums.BacktestingMetadata.ENTRIES.value: len(entries),
            common_enums.BacktestingMetadata.WINS.value: wins,
            common_enums.BacktestingMetadata.LOSES.value: len(entries) - wins,
            common_enums.BacktestingMetadata.TRADES.value: len(trades),
            common_enums.DBRows.EXCHANGES.value: exchange_names,
            common_enums.DBRows.START_TIME.value: start_time,
            common_enums.DBRows.END_TIME.value: end_time,
            common_enums.DBRows.FUTURE_CONTRACTS.value: future_contracts_by_exchange,
        }
    }


async def _get_coefficient_of_determination(exchange_managers, use_high_instead_of_end_balance):
    try:
        return numpy.average([
            await trading_api.get_coefficient_of_determination(
                exchange_manager,
                use_high_instead_of_end_balance=use_high_instead_of_end_balance
            )
            for exchange_manager in exchange_managers
        ])
    except KeyError:
        return None


def _get_markets_profitability(exchange_managers, time_frames):
    markets_profitability = {}
    min_tf = time_frame_manager.find_min_time_frame(time_frames)
    for exchange_manager in exchange_managers:
        for symbol in trading_api.get_trading_pairs(exchange_manager):
            if symbol is markets_profitability:
                continue
            try:
                markets_profitability[symbol] = \
                    backtesting.IndependentBacktesting.get_market_delta(symbol, exchange_manager, min_tf)
            except Exception:
                pass
    return markets_profitability


def _get_portfolio(exchange_managers):
    origin_portfolio = {}
    end_portfolio = {}
    for exchange_manager in exchange_managers:
        try:
            exchange_origin_portfolio = trading_api.get_origin_portfolio(exchange_manager, as_decimal=False)
        except AttributeError:
            # no initialized portfolio on this exchange
            continue
        exchange_end_portfolio = trading_api.get_portfolio(exchange_manager, as_decimal=False)
        for portfolio in (exchange_origin_portfolio, exchange_end_portfolio):
            for values in portfolio.values():
                values.pop("available", None)
        if exchange_manager.is_future:
            for position in trading_api.get_positions(exchange_manager):
                try:
                    exchange_end_portfolio[position.get_currency()]["position"] = float(position.quantity)
                except KeyError:
                    exchange_end_portfolio[position.get_currency()] = {
                        commons_constants.PORTFOLIO_AVAILABLE: 0,
                        commons_constants.PORTFOLIO_TOTAL: 0,
                        "position": float(position.quantity),
                    }

        for exchange_portfolio, portfolio in zip((exchange_origin_portfolio, exchange_end_portfolio),
                                                 (origin_portfolio, end_portfolio)):
            for currency, value_dict in exchange_portfolio.items():
                try:
                    pf_value = portfolio[currency]
                except KeyError:
                    pf_value = {}
                    portfolio[currency] = pf_value
                for key, value in value_dict.items():
                    pf_value[key] = pf_value.get(key, 0) + value
    return origin_portfolio, end_portfolio


def _get_future_contracts_by_exchange(exchange_managers):
    return {
        trading_api.get_exchange_name(exchange_manager): {
            symbol: {
                "contract_type": contract.contract_type.value,
                "position_mode": contract.position_mode.value,
                "margin_type": contract.margin_type.value
            }
            for symbol, contract in trading_api.get_pair_contracts(exchange_manager).items()
            if symbol in trading_api.get_trading_pairs(exchange_manager)
            and trading_api.is_handled_contract(contract)
        }
        for exchange_manager in exchange_managers
        if exchange_manager.is_future and hasattr(exchange_manager.exchange, "pair_contracts")
    }


async def _get_single_exchange_data(exchange_manager, trading_mode, run_start_time, user_inputs, run_dbs_identifier,
                                    is_backtesting, name):
    exchange_type = trading_enums.ExchangeTypes.FUTURE.value if exchange_manager.is_future \
        else trading_enums.ExchangeTypes.SPOT.value
    if user_inputs is None:
        user_inputs = {}
    formatted_user_inputs = {}
    for user_input in user_inputs:
        if not user_input["is_nested_config"]:
            try:
                formatted_user_inputs[user_input["tentacle"]][user_input["name"]] = user_input["value"]
            except KeyError:
                formatted_user_inputs[user_input["tentacle"]] = {
                    user_input["name"]: user_input["value"]
                }
    leverage = 0
    if exchange_manager.is_future and hasattr(exchange_manager.exchange, "get_pair_future_contract"):
        leverage = float(next(iter(trading_api.get_pair_contracts(exchange_manager).values())).current_leverage)

    backtesting_only_metadata = {
        common_enums.BacktestingMetadata.ID.value: run_dbs_identifier.backtesting_id,
        common_enums.BacktestingMetadata.OPTIMIZATION_CAMPAIGN.value: run_dbs_identifier.optimization_campaign_name,
        common_enums.BacktestingMetadata.USER_INPUTS.value: formatted_user_inputs,
    } if is_backtesting else {}
    return {
        **backtesting_only_metadata,
        **{
            common_enums.BacktestingMetadata.TIMESTAMP.value: run_start_time,
            common_enums.BacktestingMetadata.NAME.value: name or trading_mode.get_name(),
            common_enums.BacktestingMetadata.LEVERAGE.value: leverage,
            common_enums.DBRows.TRADING_TYPE.value: exchange_type,
            common_enums.DBRows.REFERENCE_MARKET.value: trading_api.get_reference_market(exchange_manager.config),
        }
    }



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from octobot.community import errors
from octobot.community.errors import (
    RequestError,
    StatusCodeRequestError,
    BotError,
    BotNotFoundError,
    NoBotDeviceError,
)
from octobot.community import models
from octobot.community.models import (
    CommunityUserAccount,
    CommunityFields,
    CommunityTentaclesPackage,
    CommunitySupports,
    CommunityDonation,
    StartupInfo,
)
from octobot.community.supabase_backend import (
    PostgresFunctions,
    SyncConfigurationStorage,
    ASyncConfigurationStorage,
    AuthenticatedAsyncSupabaseClient,
    CommunitySupabaseClient,
)

from octobot.community import community_analysis
from octobot.community import community_manager
from octobot.community import authentication
from octobot.community import graphql_requests
from octobot.community import feeds
from octobot.community import errors_upload

from octobot.community.community_analysis import (
    get_community_metrics,
    get_current_octobots_stats,
    can_read_metrics,
)
from octobot.community.community_manager import (
    CommunityManager,
)
from octobot.community.authentication import (
    CommunityAuthentication,
)
from octobot.community.graphql_requests import (
    select_startup_info_query,
    select_bot_query,
    select_bots_query,
    create_bot_query,
    create_bot_device_query,
    update_bot_config_and_stats_query,
    select_subscribed_profiles_query,
    update_bot_trades_query,
    upsert_bot_trades_query,
    update_bot_portfolio_query,
    upsert_historical_bot_portfolio_query,
)
from octobot.community.feeds import (
    AbstractFeed,
    CommunityWSFeed,
    CommunityMQTTFeed,
    community_feed_factory,
)
from octobot.community.errors_upload import (
    register_error_uploader,
    Error,
    ErrorsUploader,
)
from octobot.community.identifiers_provider import (
    IdentifiersProvider,
)

__all__ = [
    "RequestError",
    "StatusCodeRequestError",
    "BotError",
    "BotNotFoundError",
    "NoBotDeviceError",
    "IdentifiersProvider",
    "CommunityUserAccount",
    "CommunityFields",
    "get_community_metrics",
    "get_current_octobots_stats",
    "can_read_metrics",
    "CommunityManager",
    "CommunityAuthentication",
    "CommunityTentaclesPackage",
    "CommunitySupports",
    "CommunityDonation",
    "register_error_uploader",
    "Error",
    "ErrorsUploader",
    "StartupInfo",
    "PostgresFunctions",
    "SyncConfigurationStorage",
    "ASyncConfigurationStorage",
    "AuthenticatedAsyncSupabaseClient",
    "CommunitySupabaseClient",
    "select_startup_info_query",
    "select_bot_query",
    "select_bots_query",
    "create_bot_query",
    "create_bot_device_query",
    "update_bot_config_and_stats_query",
    "select_subscribed_profiles_query",
    "update_bot_trades_query",
    "upsert_bot_trades_query",
    "update_bot_portfolio_query",
    "upsert_historical_bot_portfolio_query",
    "AbstractFeed",
    "CommunityWSFeed",
    "CommunityMQTTFeed",
    "community_feed_factory",
]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot.strategy_optimizer.strategy_optimizer as strategy_optimizer
import octobot.strategy_optimizer.strategy_design_optimizer as strategy_design_optimizer
import octobot.strategy_optimizer.optimizer_settings as optimizer_settings_import
import octobot.strategy_optimizer.strategy_design_optimizer_factory as strategy_design_optimizer_factory


def create_strategy_optimizer(config, tentacles_setup_config, strategy_name) -> strategy_optimizer.StrategyOptimizer:
    return strategy_optimizer.StrategyOptimizer(config, tentacles_setup_config, strategy_name)


def create_design_strategy_optimizer(
        trading_mode, optimizer_settings, config=None, tentacles_setup_config=None
) -> strategy_design_optimizer.StrategyDesignOptimizer:
    return strategy_design_optimizer_factory.create_most_advanced_strategy_design_optimizer(
        trading_mode, config, tentacles_setup_config, optimizer_settings=optimizer_settings
    )


async def initialize_design_strategy_optimizer(strategy_optimizer, is_resuming, is_computing=False):
    strategy_optimizer.is_computing = is_computing
    return await strategy_optimizer.initialize(is_resuming)


async def update_strategy_optimizer_total_runs(optimizer, runs):
    if optimizer.empty_the_queue:
        optimizer.total_nb_runs += len(runs)


async def generate_and_save_strategy_optimizer_runs(trading_mode, tentacles_setup_config, optimizer_settings) -> list:
    optimizer = create_design_strategy_optimizer(trading_mode, optimizer_settings, None, tentacles_setup_config)
    return await optimizer.generate_and_save_run()


def create_strategy_optimizer_settings(settings_dict: dict) -> optimizer_settings_import.OptimizerSettings:
    return optimizer_settings_import.OptimizerSettings(settings_dict)


async def resume_design_strategy_optimizer(optimizer, optimizer_settings: optimizer_settings_import.OptimizerSettings):
    # continue till the queue is empty if no optimizer id is specified
    optimizer_settings.empty_the_queue = optimizer_settings.optimizer_ids is None
    optimizer_settings.optimizer_ids = optimizer_settings.optimizer_ids or await optimizer.get_queued_optimizer_ids()
    return await optimizer.resume(optimizer_settings)


def find_optimal_configuration(strategy_optimizer, TAs=None, time_frames=None, risks=None) -> None:
    strategy_optimizer.find_optimal_configuration(TAs=TAs, time_frames=time_frames, risks=risks)


def cancel_strategy_optimizer(strategy_optimizer):
    strategy_optimizer.cancel()


def print_optimizer_report(strategy_optimizer) -> None:
    strategy_optimizer.print_report()


def get_optimizer_report(strategy_optimizer) -> list:
    return strategy_optimizer.get_report()


def get_optimizer_results(strategy_optimizer) -> list:
    return strategy_optimizer.run_results


async def get_optimizer_overall_progress(strategy_optimizer) -> (int, float):
    return await strategy_optimizer.get_overall_progress()


async def get_design_strategy_optimizer_queue(trading_mode) -> list:
    return await strategy_design_optimizer.StrategyDesignOptimizer.get_run_queue(trading_mode)


async def update_design_strategy_optimizer_queue(trading_mode, updated_queue) -> list:
    return await strategy_design_optimizer.StrategyDesignOptimizer.update_run_queue(trading_mode, updated_queue)


async def is_optimizer_in_progress(strategy_optimizer) -> bool:
    return await strategy_optimizer.is_in_progress()


def is_optimizer_computing(strategy_optimizer) -> bool:
    return strategy_optimizer.is_computing


def is_optimizer_finished(strategy_optimizer) -> bool:
    return strategy_optimizer.is_finished


def get_optimizer_errors_description(strategy_optimizer) -> str:
    return strategy_optimizer.get_errors_description()


def get_optimizer_current_test_suite_progress(strategy_optimizer) -> int:
    return strategy_optimizer.get_current_test_suite_progress()


def get_optimizer_strategy(strategy_optimizer) -> object:
    return strategy_optimizer.strategy_class


def get_optimizer_all_time_frames(strategy_optimizer) -> list:
    return strategy_optimizer.all_time_frames


def get_optimizer_all_TAs(strategy_optimizer) -> list:
    return strategy_optimizer.all_TAs


def get_optimizer_all_risks(strategy_optimizer) -> list:
    return strategy_optimizer.risks


def get_optimizer_trading_mode(strategy_optimizer) -> object:
    return strategy_optimizer.trading_mode


def get_optimizer_is_properly_initialized(strategy_optimizer) -> bool:
    return strategy_optimizer.is_properly_initialized



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import abc
import os.path as path

import octobot_commons.logging as logging
import octobot_commons.tentacles_management as tentacles_management

import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_tentacles_manager.constants as tentacles_manager_constants

import octobot.constants as constants

DEFAULT_SYMBOL = "ICX/BTC"
DATA_FILE_PATH = "tests/static/"

SYMBOLS = {
    "Bitcoin": {
        "pairs": ["BTC/USDT"]
    },
    "Neo": {
        "pairs": ["NEO/BTC"]
    },
    "Ethereum": {
        "pairs": ["ETH/USDT"]
    },
    "Icon": {
        "pairs": ["ICX/BTC"]
    },
    "VeChain": {
        "pairs": ["VEN/BTC"]
    },
    "Nano": {
        "pairs": ["XRB/BTC"]
    },
    "Cardano": {
        "pairs": ["ADA/BTC"]
    },
    "Ontology": {
        "pairs": ["ONT/BTC"]
    },
    "Stellar": {
        "pairs": ["XLM/BTC"]
    },
    "Power Ledger": {
        "pairs": ["POWR/BTC"]
    },
    "Ethereum Classic": {
        "pairs": ["ETC/BTC"]
    },
    "WAX": {
        "pairs": ["WAX/BTC"]
    },
    "XRP": {
        "pairs": ["XRP/BTC"]
    },
    "Verge": {
        "pairs": ["XVG/BTC"]
    }
}

TEST_DATA_FILES_FOLDER = path.join("tests", "static")

DATA_FILES = {
    "ADA/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581774950.9324272.data"),
    "BTC/USDT": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                          "AbstractExchangeHistoryCollector_1581774962.1269426.data"),
    "ETH/USDT": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                          "AbstractExchangeHistoryCollector_1581776676.5721796.data"),
    "ICX/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581774974.669779.data"),
    "NEO/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581774982.726014.data"),
    "VEN/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581774995.2311237.data"),
    "XRB/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581775026.9255266.data"),
    "ONT/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581774988.7215023.data"),
    "XLM/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581775018.2658834.data"),
    "POWR/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                          "AbstractExchangeHistoryCollector_1581776404.9679003.data"),
    "2020ADA/USDT": path.join(TEST_DATA_FILES_FOLDER, "ExchangeHistoryDataCollector_1587769859.9278197.data"),
    "2020ADA/BTC": path.join(TEST_DATA_FILES_FOLDER, "ExchangeHistoryDataCollector_1588110698.1060486.data")
}

EXTENDED_DATA_FILES = {
    "ADA/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581775117.1713624.data"),
    "ETC/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581775133.1533682.data"),
    "NEO/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581775139.0332782.data"),
    "WAX/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581775144.480404.data"),
    "XRP/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581775149.6372743.data"),
    "XVG/BTC": path.join(constants.OPTIMIZER_DATA_FILES_FOLDER,
                         "AbstractExchangeHistoryCollector_1581775154.2598503.data")
}


class AbstractBacktestingTest:
    __metaclass__ = abc.ABCMeta

    def initialize_with_strategy(self, strategy_evaluator_class, tentacles_setup_config, config):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.config = config
        self.tentacles_setup_config = tentacles_setup_config
        self.strategy_evaluator_class = strategy_evaluator_class
        self._register_only_strategy(strategy_evaluator_class)

    # plays a backtesting market profitability:
    # 1. ICX/BTC[30m]: -13.704206241519667 % (binance_ICX_BTC_20180716_131148)
    @abc.abstractmethod
    async def test_default_run(self, strategy_tester):
        raise NotImplementedError("test_default_run not implemented")

    # plays a backtesting on a slow downtrend market:
    # 1. ICX/BTC[30m]: -13.704206241519667 % (binance_ICX_BTC_20180716_131148)
    # 2. NEO/BTC[30m]: -41.09080962800873 % (bittrex_NEO_BTC_20180722_195942)
    # 3. ONT/BTC[30m]: -17.9185520361991 % (binance_ONT_BTC_20180722_230900)
    # 4. XVG/BTC[30m]: -47.29729729729728 % (bittrex_XVG_BTC_20180726_211225)
    @abc.abstractmethod
    async def test_slow_downtrend(self, strategy_tester):
        raise NotImplementedError("test_slow_downtrend not implemented")

    # plays a backtesting on a sharp downtrend market:
    # 1. VEN/BTC[30m] -35.26645213762865 % (binance_VEN_BTC_20180716_131148)
    # 2. XRP/BTC[30m]: -47.41750358680059 (vs btc) % (bittrex_XRP_BTC_20180726_210927)
    @abc.abstractmethod
    async def test_sharp_downtrend(self, strategy_tester):
        raise NotImplementedError("test_sharp_downtrend not implemented")

    # plays a backtesting flat markets profitability:
    # 1. NEO/BTC[30m] -11.80763473053894 % (binance_NEO_BTC_20180716_131148)
    # 2. XRB/BTC[30m] -3.5209457722950255 % (binance_XRB_BTC_20180716_131148)
    # 3. ADA/BTC[30m] -6.140724946695086 % (bittrex_ADA_BTC_20180722_223357)
    # 4. WAX/BTC[30m] -8.77598152424941 % (bittrex_WAX_BTC_20180726_205032)
    @abc.abstractmethod
    async def test_flat_markets(self, strategy_tester):
        raise NotImplementedError("test_flat_markets not implemented")

    # plays a backtesting with this strategy on a slow uptrend market:
    # 1. BTC/USDT[30m]: 17.20394836443646 (vs btc) % (binance_BTC_USDT_20180428_121156)
    # 2. ADA/BTC[30m] 16.19613670133728 % (binance_ADA_BTC_20180722_223335)
    @abc.abstractmethod
    async def test_slow_uptrend(self, strategy_tester):
        raise NotImplementedError("test_slow_uptrend not implemented")

    # plays a backtesting with this strategy on a slow uptrend market:
    # 1. XLM/BTC[30m]: 30.88185223016684 (vs btc) % (binance_XLM_BTC_20180722_234305)
    # 2. POWR/BTC[30m]: 12.28597871355852 (vs btc) % (binance_POWR_BTC_20180722_234855)
    @abc.abstractmethod
    async def test_sharp_uptrend(self, strategy_tester):
        raise NotImplementedError("test_sharp_uptrend not implemented")

    # plays a backtesting with this strategy on a slow uptrend market followed by a slow downtrend market:
    # 1. ETC/BTC[30m] -6.428386403538222 % (bittrex_ETC_BTC_20180726_210341)
    @abc.abstractmethod
    async def test_up_then_down(self, strategy_tester):
        raise NotImplementedError("test_up_then_down not implemented")

    @abc.abstractmethod
    def _handle_results(self, independent_backtesting, profitability):
        raise NotImplementedError("_handle_results not implemented")

    @abc.abstractmethod
    async def _run_backtesting_with_current_config(self, data_file_to_use):
        raise NotImplementedError("_run_backtesting_with_current_config not implemented")

    async def _run_and_handle_results(self, data_file, expected_profitability):
        independent_backtesting = None
        try:
            independent_backtesting = await self._run_backtesting_with_current_config(data_file)
            self._handle_results(independent_backtesting, expected_profitability)
        finally:
            if independent_backtesting is not None:
                await independent_backtesting.stop()

    async def run_test_default_run(self, profitability):
        await self._run_and_handle_results(DATA_FILES[DEFAULT_SYMBOL], profitability)

    async def run_test_slow_downtrend(self, profitability_1, profitability_2, profitability_3, profitability_4,
                                      skip_extended=False):
        await self._run_and_handle_results(DATA_FILES["ICX/BTC"], profitability_1)
        await self._run_and_handle_results(DATA_FILES["ONT/BTC"], profitability_2)
        if not skip_extended:
            await self._run_and_handle_results(EXTENDED_DATA_FILES["NEO/BTC"], profitability_3)
            await self._run_and_handle_results(EXTENDED_DATA_FILES["XVG/BTC"], profitability_4)

    async def run_test_sharp_downtrend(self, profitability_1, profitability_2, skip_extended=False):
        await self._run_and_handle_results(DATA_FILES["VEN/BTC"], profitability_1)
        if not skip_extended:
            await self._run_and_handle_results(EXTENDED_DATA_FILES["XRP/BTC"], profitability_2)

    async def run_test_flat_markets(self, profitability_1, profitability_2, profitability_3, profitability_4,
                                    skip_extended=False):
        await self._run_and_handle_results(DATA_FILES["NEO/BTC"], profitability_1)
        await self._run_and_handle_results(DATA_FILES["XRB/BTC"], profitability_2)
        if not skip_extended:
            await self._run_and_handle_results(EXTENDED_DATA_FILES["ADA/BTC"], profitability_3)
            await self._run_and_handle_results(EXTENDED_DATA_FILES["WAX/BTC"], profitability_4)

    async def run_test_slow_uptrend(self, profitability_1, profitability_2):
        await self._run_and_handle_results(DATA_FILES["BTC/USDT"], profitability_1)
        await self._run_and_handle_results(DATA_FILES["ADA/BTC"], profitability_2)

    async def run_test_sharp_uptrend(self, profitability_1, profitability_2):
        await self._run_and_handle_results(DATA_FILES["XLM/BTC"], profitability_1)
        await self._run_and_handle_results(DATA_FILES["POWR/BTC"], profitability_2)

    async def run_test_up_then_down(self, profitability_1, skip_extended=False):
        if not skip_extended:
            await self._run_and_handle_results(EXTENDED_DATA_FILES["ETC/BTC"], profitability_1)

    def _register_only_strategy(self, strategy_evaluator_class):
        try:
            import tentacles.Evaluator as Evaluator
            import octobot_evaluators.evaluators as evaluators
            tentacles_activation = tentacles_manager_api.get_tentacles_activation(self.tentacles_setup_config)
            for evaluator_name in tentacles_activation[tentacles_manager_constants.TENTACLES_EVALUATOR_PATH]:
                if tentacles_management.get_class_from_string(
                        evaluator_name,
                        evaluators.StrategyEvaluator,
                        Evaluator.Strategies,
                        tentacles_management.evaluator_parent_inspection) is not None:
                    tentacles_activation[tentacles_manager_constants.TENTACLES_EVALUATOR_PATH][evaluator_name] = False
            tentacles_activation[tentacles_manager_constants.TENTACLES_EVALUATOR_PATH][
                strategy_evaluator_class.get_name()] = True
        except ImportError as e:
            self.logger.error(f"Backtesting tests requires OctoBot-Evaluator package installed {e}")



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import contextlib

import pytest
import pytest_asyncio
import mock
import requests
import aiohttp

import octobot.community as community
import octobot.constants as constants
import octobot_commons.authentication as authentication
import octobot_commons.configuration

AUTH_URL = "https://oh.fake/auth"
AUTH_RETURN = {
    "access_token": "1",
    "refresh_token": "2",
    "expires_in": 3600,
}
EMAIL_RETURN = {
    "email": "plop"
}
AUTH_HEADER_RETURN = {
    community.CommunityAuthentication.SESSION_HEADER: "helloooo",
}


class MockedResponse:
    def __init__(self, status_code=200, json=None, headers={}):
        self.status_code = status_code
        self.json_resp = json
        self.headers = headers

    def json(self):
        return self.json_resp


@pytest.fixture
def auth():
    community.IdentifiersProvider.use_production()
    authenticator = community.CommunityAuthentication(AUTH_URL, None)
    authenticator.supabase_client = mock.Mock(
        sign_in=mock.AsyncMock(),
        sign_in_with_otp_token=mock.AsyncMock(),
        sign_out=mock.Mock(),
        auth=mock.Mock(_storage_key="_storage_key"),
        close=mock.AsyncMock(),
    )
    return authenticator


@pytest_asyncio.fixture
async def logged_in_auth(auth):
    auth.user_account.has_user_data = mock.Mock(return_value=True)
    auth.user_account.get_email = mock.Mock(return_value="plop")
    return auth


def test_constructor():
    with mock.patch.object(community.CommunityAuthentication, "login", mock.Mock()) as login_mock:
        community.IdentifiersProvider.use_production()
        community.CommunityAuthentication(AUTH_URL, None)
        auth = community.CommunityAuthentication(AUTH_URL, None)
        login_mock.assert_not_called()
        assert not auth.user_account.supports.is_supporting()
        assert auth.initialized_event is None


@pytest.mark.asyncio
async def test_login(auth):
    resp_mock = mock.Mock()
    with mock.patch.object(community.CommunityAuthentication, "_reset_tokens", mock.Mock()) as reset_mock, \
            mock.patch.object(community.CommunityAuthentication, "_ensure_community_url", mock.Mock()) \
            as _ensure_community_url_mock, \
            mock.patch.object(community.CommunityAuthentication, "_ensure_email", mock.Mock()) \
            as _ensure_email_mock, \
            mock.patch.object(community.CommunityAuthentication, "_on_account_updated", mock.AsyncMock()) \
            as _on_account_updated_mock, \
            mock.patch.object(community.CommunityAuthentication, "is_logged_in", mock.Mock()) \
            as is_logged_in_mock, \
            mock.patch.object(community.CommunityAuthentication, "on_signed_in", mock.AsyncMock()) \
            as on_signed_in_mock:
        await auth.login("username", "password")
        reset_mock.assert_called_once()
        _ensure_community_url_mock.assert_called_once()
        _ensure_email_mock.assert_called_once()
        _on_account_updated_mock.assert_called_once()
        is_logged_in_mock.assert_called_once()
        on_signed_in_mock.assert_called_once()
        auth.supabase_client.sign_in.assert_awaited_once_with("username", "password")
        auth.supabase_client.sign_in_with_otp_token.assert_not_called()
        auth.supabase_client.sign_in.reset_mock()
        await auth.login(None, None, password_token="password_t")
        auth.supabase_client.sign_in.assert_not_called()
        auth.supabase_client.sign_in_with_otp_token.assert_awaited_once_with("password_t")

def test_logout(auth):
    with mock.patch.object(community.CommunityAuthentication, "_reset_tokens", mock.Mock()) as reset_mock, \
         mock.patch.object(community.CommunityAuthentication, "remove_login_detail", mock.Mock()) as remove_mock:
        auth.logout()
        reset_mock.assert_called_once()
        remove_mock.assert_called_once()
        auth.supabase_client.sign_out.assert_called_once()


def test_get_logged_in_email_authenticated(logged_in_auth):
    assert logged_in_auth.get_logged_in_email() == "plop"


def test_get_logged_in_email_unauthenticated(auth):
    with pytest.raises(authentication.AuthenticationRequired):
        auth.get_logged_in_email()


def test_can_authenticate(auth):
    assert auth.can_authenticate() is True


def test_ensure_community_url(auth):
    with mock.patch.object(auth, "can_authenticate", mock.Mock(return_value=False)) as can_authenticate_mock:
        with pytest.raises(authentication.UnavailableError):
            auth._ensure_community_url()
        can_authenticate_mock.assert_called_once()
    with mock.patch.object(auth, "can_authenticate", mock.Mock(return_value=True)) as can_authenticate_mock:
        auth._ensure_community_url()
        can_authenticate_mock.assert_called_once()


def test_is_logged_in(auth):
    auth.user_account.has_user_data = mock.Mock(return_value=False)
    assert auth.is_logged_in() is False
    auth.supabase_client.is_signed_in.assert_called_once()
    auth.user_account.has_user_data.assert_called_once()
    auth.user_account.has_user_data = mock.Mock(return_value=True)
    assert auth.is_logged_in() is True


def test_remove_login_detail(auth):
    with mock.patch.object(auth, "_reset_login_token", mock.Mock()) as _reset_login_token_mock, \
            mock.patch.object(auth, "_save_bot_id", mock.Mock()) as _save_bot_id_mock:
        auth.remove_login_detail()
        _reset_login_token_mock.assert_called_once()
        _save_bot_id_mock.assert_called_once()


def test_reset_login_token(auth):
    with mock.patch.object(octobot_commons.configuration.Configuration, "save", mock.Mock()) as save_mock:
        auth.configuration_storage.configuration = octobot_commons.configuration.Configuration("", "")
        auth.configuration_storage.configuration.config = {
            constants.CONFIG_COMMUNITY: {
                "_storage_key": "plop"
            }
        }
        auth._reset_login_token()
        assert auth.configuration_storage.configuration.config[constants.CONFIG_COMMUNITY]["_storage_key"] == ""
        save_mock.assert_called_once_with()


def test_get_saved_bot_id(auth):
    assert auth._get_saved_bot_id() is None
    auth.configuration_storage.configuration = octobot_commons.configuration.Configuration("", "")
    auth.configuration_storage.configuration.config = {
        constants.CONFIG_COMMUNITY: {
            constants.CONFIG_COMMUNITY_BOT_ID: "bid"
        }
    }
    assert auth._get_saved_bot_id() == "bid"


def test_authenticated(auth):
    @authentication.authenticated
    def mock_func(*_):
        pass
    with mock.patch.object(auth, "ensure_token_validity", mock.Mock()) as ensure_token_validity_mock:
        mock_func(auth)
        ensure_token_validity_mock.assert_called_once()


def test_update_supports(auth):
    with mock.patch.object(community.CommunitySupports, "from_community_dict", mock.Mock()) as from_community_dict_mock:
        auth._update_supports(400, {})
        from_community_dict_mock.assert_not_called()
        auth._update_supports(200, {})
        from_community_dict_mock.assert_called_once_with({})


# TODO restore test when implemented
@pytest.mark.asyncio
async def _test_auth_and_fetch_supports(auth):
    with mock.patch.object(auth, "_async_try_auto_login", mock.AsyncMock()) as _async_try_auto_login_mock:
        with mock.patch.object(auth, "is_logged_in", mock.Mock(return_value=False)) as is_logged_in_mock:
            auth.initialized_event = asyncio.Event()
            await auth._initialize_account()
            _async_try_auto_login_mock.assert_called_once()
            is_logged_in_mock.assert_called_once()
            assert auth.initialized_event.is_set()
            _async_try_auto_login_mock.reset_mock()
        with mock.patch.object(auth, "is_logged_in", mock.Mock(return_value=True)) as is_logged_in_mock, \
                mock.patch.object(auth, "_update_supports", mock.Mock()) as _update_supports_mock:
            resp_mock = mock.AsyncMock()
            resp_mock.status = 200
            resp_mock.json = mock.AsyncMock(return_value="plop")

            @contextlib.asynccontextmanager
            async def async_get(*_, **__):
                yield resp_mock

            auth._aiohttp_gql_session = mock.AsyncMock()
            auth._aiohttp_gql_session.get = async_get
            auth.initialized_event = asyncio.Event()
            await auth._initialize_account()
            _async_try_auto_login_mock.assert_called_once()
            is_logged_in_mock.assert_called_once()
            _update_supports_mock.assert_called_once_with(200, "plop")
            assert auth.initialized_event.is_set()


# TODO restore test when implemented
def _test_public_update_supports(auth):
    with mock.patch.object(auth, "get", mock.Mock(return_value=MockedResponse(status_code=200, json=AUTH_RETURN))) \
        as get_mock, \
         mock.patch.object(auth, "_update_supports", mock.Mock()) as _update_supports_mock:
        auth.update_supports()
        get_mock.assert_called_once()
        _update_supports_mock.assert_called_once_with(200, AUTH_RETURN)


def test_is_initialized(auth):
    assert auth.is_initialized() is False
    auth.initialized_event = asyncio.Event()
    assert auth.is_initialized() is False
    auth.initialized_event.set()
    assert auth.is_initialized() is True


def test_init_account(auth):
    with mock.patch.object(asyncio, "create_task", mock.Mock(return_value="task")) as create_task_mock, \
            mock.patch.object(auth, "_initialize_account", mock.Mock(return_value="coro")) \
            as _auth_and_fetch_account_mock:
        auth.init_account()
        create_task_mock.assert_called_once_with("coro")
        _auth_and_fetch_account_mock.assert_called_once()
        assert auth._fetch_account_task == "task"


@pytest.mark.asyncio
async def test_stop(auth):
    auth._fetch_account_task = mock.Mock()
    auth._fetch_account_task.cancel = mock.Mock()
    auth._fetch_account_task.done = mock.Mock(return_value=True)
    await auth.stop()
    auth.supabase_client.close.assert_awaited_once()
    auth.supabase_client.close.reset_mock()
    auth._fetch_account_task.cancel.assert_not_called()
    auth._fetch_account_task.done = mock.Mock(return_value=False)

    await auth.stop()
    auth.supabase_client.close.assert_awaited_once()
    auth._fetch_account_task.cancel.assert_called_once()

    auth.supabase_client.close.reset_mock()
    auth._fetch_account_task.cancel.reset_mock()
    await auth.stop()
    auth.supabase_client.close.assert_awaited_once()
    auth._fetch_account_task.cancel.assert_called_once()
